import shutil
import typing as t
from pathlib import Path

import typer

import arguebuf as ag
from arguebuf.cli.translator import Translator
from arguebuf.converters.to_graphviz import to_graphviz
from arguebuf.schemas.graphviz import EdgeStyle

from . import model

cli = typer.Typer()


@cli.command()
def translate(
    input_folder: Path,
    source_lang: str,
    target_lang: str,
    auth_key: str,
    input_glob: str,
    output_folder: t.Optional[Path] = None,
    output_format: ag.GraphFormat = ag.GraphFormat.ARGUEBUF,
    clean: bool = False,
    overwrite: bool = False,
    start: int = 1,
) -> None:
    if not output_folder:
        output_folder = input_folder

    if clean:
        shutil.rmtree(output_folder)
        output_folder.mkdir()

    path_pairs = model.PathPair.create(input_folder, output_folder, input_glob, ".json")
    translator = Translator(auth_key, source_lang, target_lang)
    bar: t.Iterable[model.PathPair]

    with typer.progressbar(
        path_pairs[start - 1 :],
        item_show_func=model.PathPair.label,
        show_pos=True,
    ) as bar:
        for path_pair in bar:
            if overwrite or not path_pair.target.exists():
                graph = ag.from_file(path_pair.source)
                translator.translate(graph)
                ag.to_file(graph, path_pair.target, output_format)


def _strip_node_labels(node: ag.AbstractNode) -> str:
    label: str = ""

    for char in node.label:
        if char.isspace():
            label += char
        else:
            label += "–"

    return label


@cli.command()
def render(
    input_folder: Path,
    input_glob: str,
    output_folder: t.Optional[Path] = None,
    output_format: str = ".pdf",
    strip_scheme_nodes: bool = False,
    strip_node_labels: bool = False,
    edge_style: t.Optional[EdgeStyle] = None,
    nodesep: t.Optional[float] = None,
    ranksep: t.Optional[float] = None,
    node_wrap_col: t.Optional[int] = None,
    node_margin: t.Tuple[float, float] = (0, 0),
    font_name: t.Optional[str] = None,
    font_size: t.Optional[float] = None,
    clean: bool = False,
    overwrite: bool = False,
    start: int = 1,
    max_nodes: t.Optional[int] = None,
) -> None:
    if not output_folder:
        output_folder = input_folder

    if clean:
        shutil.rmtree(output_folder)
        output_folder.mkdir()

    paths = model.PathPair.create(
        input_folder, output_folder, input_glob, output_format
    )
    bar: t.Iterable[model.PathPair]

    with typer.progressbar(
        paths[start - 1 :],
        item_show_func=model.PathPair.label,
        show_pos=True,
    ) as bar:
        for path_pair in bar:
            if overwrite or not path_pair.target.exists():
                g = ag.from_file(path_pair.source)

                if strip_scheme_nodes:
                    g.strip_scheme_nodes()

                gv = to_graphviz(
                    g,
                    nodesep=nodesep,
                    ranksep=ranksep,
                    wrap_col=node_wrap_col,
                    margin=node_margin
                    if all(margin > 0 for margin in node_margin)
                    else None,
                    font_name=font_name,
                    font_size=font_size,
                    atom_label=_strip_node_labels if strip_node_labels else None,
                    scheme_label=_strip_node_labels if strip_node_labels else None,
                    edge_style=edge_style,
                    max_nodes=max_nodes,
                )
                ag.render(gv, path_pair.target)


@cli.command()
def convert(
    input_folder: Path,
    input_glob: str,
    output_folder: t.Optional[Path] = None,
    output_format: ag.GraphFormat = ag.GraphFormat.ARGUEBUF,
    clean: bool = False,
    overwrite: bool = False,
    start: int = 1,
    text_folder: t.Optional[Path] = None,
    text_suffix: str = ".txt",
) -> None:
    if not output_folder:
        output_folder = input_folder

    if clean:
        shutil.rmtree(output_folder)
        output_folder.mkdir()

    paths = model.PathPair.create(input_folder, output_folder, input_glob, ".json")
    bar: t.Iterable[model.PathPair]

    with typer.progressbar(
        paths[start - 1 :],
        item_show_func=model.PathPair.label,
        show_pos=True,
    ) as bar:
        for path_pair in bar:
            if overwrite or not path_pair.target.exists():
                text_file = None

                if text_folder:
                    text_file = text_folder / path_pair.source.relative_to(
                        input_folder
                    ).with_suffix(text_suffix)

                graph = ag.from_file(path_pair.source, text_file=text_file)

                ag.to_file(graph, path_pair.target, output_format)


@cli.command()
def statistics(
    input_folder: Path,
    input_glob: str,
):
    files = sorted(input_folder.glob(input_glob))

    graphs = [ag.from_file(file) for file in files]
    atom_nodes = [len(graph.atom_nodes) for graph in graphs]
    scheme_nodes = [len(graph.scheme_nodes) for graph in graphs]
    edges = [len(graph.edges) for graph in graphs]

    total_graphs = len(graphs)
    total_atom_nodes = sum(atom_nodes)
    total_scheme_nodes = sum(scheme_nodes)
    total_edges = sum(edges)

    typer.echo(
        f"""Total Graphs: {total_graphs}

Total Atom Nodes: {total_atom_nodes}
Total Scheme Nodes: {total_scheme_nodes}
Total Edges: {total_edges}

Atom Nodes per Graph: {total_atom_nodes / total_graphs}
Scheme Nodes per Graph: {total_scheme_nodes / total_graphs}
Edges per Graph: {total_edges / total_graphs}

Max. Atom Nodes: {max(atom_nodes)}
Max. Scheme Nodes: {max(scheme_nodes)}
Max. Edges: {max(edges)}

Min. Atom Nodes: {min(atom_nodes)}
Min. Scheme Nodes: {min(scheme_nodes)}
Min. Edges: {min(edges)}"""
    )



from __future__ import annotations

import typing as t
from dataclasses import dataclass, field

from pendulum.datetime import DateTime

from arguebuf.models.metadata import Metadata
from arguebuf.models.userdata import Userdata
from arguebuf.services import utils


@dataclass()
class Resource:
    text: t.Any
    title: t.Optional[str] = None
    source: t.Optional[str] = None
    timestamp: t.Optional[DateTime] = None
    metadata: Metadata = field(default_factory=Metadata)
    userdata: Userdata = field(default_factory=dict)
    _id: str = field(default_factory=utils.uuid)

    @property
    def id(self) -> str:
        return self._id

    @property
    def plain_text(self) -> str:
        """Generate a string from Resource object."""
        return utils.xstr(self.text)



from __future__ import annotations

import json
import typing as t

from arguebuf.converters.config import GraphFormat
from arguebuf.converters.to_dict import to_dict
from arguebuf.models.graph import Graph


def to_json(
    graph: Graph,
    obj: t.TextIO,
    format: GraphFormat = GraphFormat.ARGUEBUF,
    pretty: bool = False,
) -> None:
    """Export structure of Graph instance to JSON argument graph format."""
    json.dump(
        to_dict(graph, format),
        obj,
        ensure_ascii=False,
        indent=4 if pretty else None,
    )



from __future__ import annotations

import csv
import typing as t

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.models.graph import Graph
from arguebuf.models.scheme import Attack, Support
from arguebuf.services import utils


def from_brat(
    obj: t.TextIO,
    name: t.Optional[str] = None,
    config: ConverterConfig = DefaultConverter,
) -> Graph:
    """Generate Graph structure from BRAT argument graph file (reference: https://brat.nlplab.org/)"""
    reader = csv.reader(obj, delimiter="\t")
    g = config.GraphClass(name)

    atom_nodes = {}
    mc = config.AtomNodeClass(utils.parse("", config.nlp))
    g.add_node(mc)
    g._major_claim = mc

    for row in reader:
        userdata = row[1].split()

        if row[0].startswith("T"):
            if userdata[0] == "MajorClaim":
                mc.text = utils.parse(f"{mc.plain_text}. {row[2]}", config.nlp)
            else:
                atom = config.AtomNodeClass(utils.parse(row[2], config.nlp))
                g.add_node(atom)
                atom_nodes[row[0]] = atom

        elif row[0].startswith("A") or row[0].startswith("R"):
            if row[0].startswith("A"):
                scheme_type = (
                    Attack.DEFAULT if userdata[2] == "Against" else Support.DEFAULT
                )
                source = atom_nodes[userdata[1]]
                target = mc
            else:
                scheme_type = (
                    Attack.DEFAULT if userdata[0] == "attacks" else Support.DEFAULT
                )
                source = atom_nodes[userdata[1].split(":")[1]]
                target = atom_nodes[userdata[2].split(":")[1]]

            scheme = config.SchemeNodeClass(scheme_type)
            g.add_node(scheme)

            g.add_edge(config.EdgeClass(source, scheme))
            g.add_edge(config.EdgeClass(scheme, target))

    return g



import textwrap
import typing as t
from pathlib import Path

from graphviz import ENGINES, FORMATS, Digraph

from arguebuf.models.edge import Edge
from arguebuf.models.graph import Graph
from arguebuf.models.node import AtomNode, SchemeNode
from arguebuf.schemas.graphviz import EdgeStyle, GraphvizGraph
from arguebuf.services.utils import uuid

try:
    from pygraphviz import AGraph

    init_gv_graph = lambda *args, **kwargs: AGraph(*args, **kwargs, directed=True)
    add_gv_node = lambda graph, *args, **kwargs: graph.add_node(*args, **kwargs)
    add_gv_edge = lambda graph, *args, **kwargs: graph.add_edge(*args, **kwargs)
except ModuleNotFoundError:
    init_gv_graph = lambda *args, **kwargs: Digraph(*args, **kwargs)
    add_gv_node = lambda graph, *args, **kwargs: graph.node(*args, **kwargs)
    add_gv_edge = lambda graph, *args, **kwargs: graph.edge(*args, **kwargs)


def to_graphviz(
    graph: Graph,
    nodesep: t.Optional[float] = None,
    ranksep: t.Optional[float] = None,
    wrap_col: t.Optional[int] = None,
    margin: t.Optional[t.Tuple[float, float]] = None,
    font_name: t.Optional[str] = None,
    font_size: t.Optional[float] = None,
    atom_label: t.Optional[t.Callable[[AtomNode], str]] = None,
    scheme_label: t.Optional[t.Callable[[SchemeNode], str]] = None,
    graph_attr: t.Optional[t.Mapping[str, str]] = None,
    node_attr: t.Optional[t.Mapping[str, str]] = None,
    edge_attr: t.Optional[t.Mapping[str, str]] = None,
    edge_style: t.Optional[EdgeStyle] = None,
    max_nodes: t.Optional[int] = None,
) -> t.Optional[GraphvizGraph]:
    """Transform a Graph instance into an instance of GraphViz directed graph. Make sure that a GraphViz Executable path is set on your machine for visualization. Refer to the GraphViz library for additional information."""

    if len(graph.nodes) > (max_nodes or 1000):
        return None

    gv_margin: t.Callable[[t.Tuple[float, float]], str] = lambda x: f"{x[0]},{x[1]}"

    if not graph_attr:
        graph_attr = {}

    if not node_attr:
        node_attr = {}

    if not edge_attr:
        edge_attr = {}

    gv_graph = init_gv_graph(
        name=str(graph.name),
        strict=True,
    )
    # NameError

    gv_graph.node_attr.update(
        {
            "fontname": font_name or "Arial",
            "fontsize": str(font_size or 11),
            "margin": gv_margin(margin or (0.15, 0.1)),
            "style": "rounded,filled",
            "shape": "box",
            "width": "0",
            "height": "0",
            **node_attr,
        }
    )
    gv_graph.edge_attr.update({"color": "#9E9E9E", **edge_attr})
    gv_graph.graph_attr.update(
        {
            # "bgcolor": "#000000",
            "rankdir": "BT",
            "margin": "0",
            "nodesep": str(nodesep or 0.25),
            "ranksep": str(ranksep or 0.5),
            "overlap": "false",
            "splines": edge_style.value if edge_style else EdgeStyle.STEP.value,
            **graph_attr,
        }
    )

    for node in graph._atom_nodes.values():
        _atom_to_gv(
            node,
            gv_graph,
            graph.major_claim == node,
            label_func=atom_label,
            wrap_col=wrap_col or 36,
        )

    for node in graph._scheme_nodes.values():
        _scheme_to_gv(
            node,
            gv_graph,
            label_func=scheme_label,
        )

    for edge in graph._edges.values():
        _edge_to_gv(edge, gv_graph)

    return gv_graph


def _atom_to_gv(
    node: AtomNode,
    g: GraphvizGraph,
    major_claim: bool,
    wrap_col: int,
    label_func: t.Optional[t.Callable[[AtomNode], str]] = None,
) -> None:
    """Submethod used to export Graph object g into GV Graph format."""
    color = node.color(major_claim)
    label = label_func(node) if label_func else node.label

    # TODO: Improve wrapping
    # https://stackoverflow.com/a/26538082/7626878
    add_gv_node(
        g,
        node.id,
        label=textwrap.fill(label, wrap_col).strip(),
        fontcolor=color.fg,
        fillcolor=color.bg,
        color=color.border,
        root=str(major_claim),
    )


def _scheme_to_gv(
    node: SchemeNode,
    g: GraphvizGraph,
    label_func: t.Optional[t.Callable[[SchemeNode], str]] = None,
) -> None:
    """Submethod used to export Graph object g into GV Graph format."""

    color = node.color(major_claim=False)
    label = label_func(node) if label_func else node.label

    add_gv_node(
        g,
        node.id,
        label=label,
        fontcolor=color.fg,
        fillcolor=color.bg,
        color=color.border,
    )


def _edge_to_gv(edge: Edge, g: GraphvizGraph) -> None:
    """Submethod used to export Graph object g into GV Graph format."""
    add_gv_edge(
        g,
        edge.source.id,
        edge.target.id,
    )


def render(
    graph: GraphvizGraph,
    path: t.Union[Path, str],
    prog: str = "dot",
) -> None:
    """Visualize a Graph instance using a GraphViz backend. Make sure that a GraphViz Executable path is set on your machine for visualization."""
    if isinstance(path, str):
        path = Path(path)

    if path.suffix.removeprefix(".") not in FORMATS:
        raise ValueError(
            f"You need to provide a path with a file ending supported by graphviz: {FORMATS}"
        )

    if prog not in ENGINES:
        raise ValueError(
            f"You need to provide a prog that is supported by graphviz: {ENGINES}"
        )

    if isinstance(graph, Graph):
        raise ValueError(
            "This method expects a graph in the 'DOT' format."
            "Please use 'arguebuf.to_graphviz(graph)' to convert your argument graph to the 'DOT' format."
        )
    elif isinstance(graph, Digraph):
        graph.render(outfile=path, engine=prog)
    else:
        graph.draw(path=path, prog=prog)



from __future__ import annotations

import typing as t


from arguebuf.models.resource import Resource
from arguebuf.services import utils


class Reference:
    _resource: t.Optional[Resource]
    offset: t.Optional[int]
    text: t.Any

    def __init__(
        self,
        resource: t.Optional[Resource] = None,
        offset: t.Optional[int] = None,
        text: t.Optional[t.Any] = None,
    ) -> None:
        self._resource = resource
        self.offset = offset
        self.text = text

    @property
    def plain_text(self) -> str:
        """Generate a string from Resource object."""
        return utils.xstr(self.text)

    @property
    def resource(self) -> t.Optional[Resource]:
        return self._resource





from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.converters.from_protobuf import from_protobuf
from arguebuf.converters.to_protobuf import to_protobuf
from arguebuf.models.graph import Graph


def copy(obj: Graph, config: ConverterConfig = DefaultConverter) -> Graph:
    """Contents of Graph instance are copied into new Graph object."""

    return from_protobuf(to_protobuf(obj), obj.name, config)



from __future__ import annotations

import typing as t
from pathlib import Path

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.converters.from_io import from_io
from arguebuf.models.graph import Graph
from arguebuf.models.resource import Resource


def from_file(
    file: t.Union[Path, str],
    text_file: t.Optional[t.Union[Path, str]] = None,
    config: ConverterConfig = DefaultConverter,
) -> Graph:
    """Generate Graph structure from a File."""
    if isinstance(file, str):
        file = Path(file)

    with file.open("r", encoding="utf-8") as fp:
        graph = from_io(fp, file.suffix, file.stem, config)

    if not text_file:
        text_file = file.with_suffix(".txt")
    elif isinstance(text_file, str):
        text_file = Path(text_file)

    if text_file.exists():
        text = text_file.read_text()
        graph.add_resource(Resource(text))

    return graph


def from_folder(
    folder: t.Union[Path, str],
    pattern: str,
    text_folder: t.Optional[t.Union[Path, str]] = None,
    text_suffix: str = ".txt",
    config: ConverterConfig = DefaultConverter,
) -> t.Dict[Path, Graph]:
    """Load all graphs matching the specified `pattern` in `path`.

    Args:
        path: Folder containing the graphs to be loaded.
        pattern: Unix glob pattern to filter the available files.
            Recursive matching can be achieved by prepending `**/` to any pattern.
            For instance, all `json` files of a folder can be retrieved with `**/*.json`.
            Supports the following wildcards: <https://docs.python.org/3/library/fnmatch.html#module-fnmatch>
        atom_class: Allows to override the class used for atom nodes in case a specialized subclass has been created. Defaults to `AtomNode`.
        scheme_class: Allows to override the class used for scheme nodes in case a specialized subclass has been created. Defaults to `SchemeNode`.
        edge_class: Allows to override the class used for edges in case a specialized subclass has been created. Defaults to `Edge`.
        nlp: Optionally pass a function to transforms all texts of atom nodes and resources to arbitrary Python objects.
            Useful when using `spacy` to generate embeddings.
            In this case, you can load a model with `spacy.load(...)` and pass the resulting `nlp` function via this parameter.

    Returns:
        Dictionary containing all found file paths as well as the loaded graphs.
    """

    if isinstance(folder, str):
        folder = Path(folder)

    if isinstance(text_folder, str):
        text_folder = Path(text_folder)

    graphs: dict[Path, Graph] = {}

    for file in sorted(folder.glob(pattern)):
        text_file = None

        if text_folder:
            text_file = text_folder / file.relative_to(folder).with_suffix(text_suffix)

        graphs[file] = from_file(file, text_file, config)

    return graphs



from __future__ import annotations

import typing as t
from dataclasses import dataclass
from pathlib import Path


@dataclass
class PathPair:
    """A pair of paths for modification of files."""

    source: Path
    target: Path

    @classmethod
    def create(
        cls,
        path_in: Path,
        path_out: Path,
        input_glob: t.Optional[str],
        output_suffix: t.Optional[str],
    ) -> t.List[PathPair]:
        pairs: list[PathPair] = []

        if path_in.is_file():
            if path_out.suffix:
                pairs.append(cls(path_in, path_out))
            elif output_suffix:
                pairs.append(cls(path_in, path_out.with_suffix(output_suffix)))
            else:
                raise ValueError(
                    "File given as input. Please also provide a file as output."
                )

        elif path_in.is_dir() and input_glob and output_suffix:
            path_out.mkdir(parents=True, exist_ok=True)

            files_in = sorted(path_in.glob(input_glob))
            files_out: list[Path] = []

            for file_in in files_in:
                file_out = path_out / file_in.relative_to(path_in)
                file_out = file_out.with_suffix(output_suffix)
                file_out.parent.mkdir(parents=True, exist_ok=True)

                files_out.append(file_out)

            pairs.extend(
                cls(file_in, file_out) for file_in, file_out in zip(files_in, files_out)
            )

        else:
            raise ValueError(
                "Folder given as input. Please also provide a folder as output together with a shell globbing pattern and an output suffix."
            )

        return pairs

    @staticmethod
    def label(path_pair: t.Optional[PathPair]) -> str:
        """Generate a string for representing a path pair.

        Args:
            path_pair: The item that should be represented.

        Returns:
            A label for use in UI contexts.
        """

        return path_pair.source.name if path_pair else ""



import typing as t

SchemeType = t.Literal["RA", "CA", "MA", "PA", ""]
NodeType = t.Literal["RA", "CA", "MA", "PA", "", "I", "TA", "YA", "L"]
DATE_FORMAT = "YYYY-MM-DD HH:mm:ss"


class Node(t.TypedDict):
    nodeID: str
    text: str
    type: NodeType
    timestamp: str


class Edge(t.TypedDict):
    edgeID: str
    fromID: str
    toID: str
    formEdgeID: None


class Locution(t.TypedDict):
    nodeID: str
    personID: str
    timestamp: str
    start: t.Optional[str]
    end: t.Optional[str]
    source: t.Optional[str]


class Graph(t.TypedDict):
    nodes: t.List[Node]
    edges: t.List[Edge]
    locutions: t.List[Locution]



from enum import Enum


class EdgeType(Enum):
    SEGMENTATION = "seg"
    SUPPORT_DEFAULT = "sup"
    SUPPORT_EXAMPLE = "exa"
    ADDITIONAL_SOURCE = "add"
    ATTACK_ATOM = "reb"
    ATTACK_SCHEME = "und"



from __future__ import annotations

import importlib.metadata

from arg_services.graph.v1 import graph_pb2

from arguebuf.models.analyst import Analyst
from arguebuf.models.edge import Edge
from arguebuf.models.graph import Graph
from arguebuf.models.metadata import Metadata
from arguebuf.models.node import AbstractNode, AtomNode, SchemeNode
from arguebuf.models.participant import Participant
from arguebuf.models.reference import Reference
from arguebuf.models.resource import Resource
from arguebuf.models.scheme import (
    Attack,
    Preference,
    Rephrase,
    Support,
    attack2protobuf,
    preference2protobuf,
    rephrase2protobuf,
    support2protobuf,
)
from arguebuf.services import dt


def to_protobuf(obj: Graph) -> graph_pb2.Graph:
    """Export structure of Graph instance to PROTOBUF argument graph format."""
    try:
        version = importlib.metadata.version("arg_services")
    except importlib.metadata.PackageNotFoundError:
        version = ""

    g = graph_pb2.Graph(
        schema_version=1,
        library_version=version,
        metadata=metadata_to_protobuf(obj.metadata),
    )

    for node_id, node in obj._nodes.items():
        g.nodes[node_id].CopyFrom(node_to_protobuf(node))

    for edge_id, edge in obj._edges.items():
        g.edges[edge_id].CopyFrom(edge_to_protobuf(edge))

    if obj._major_claim:
        g.major_claim = obj._major_claim.id

    for resource_id, resource in obj._resources.items():
        g.resources[resource_id].CopyFrom(resource_to_protobuf(resource))

    for participant_id, participant in obj._participants.items():
        g.participants[participant_id].CopyFrom(participant_to_protobuf(participant))

    for analyst_id, analyst in obj._analysts.items():
        g.analysts[analyst_id].CopyFrom(analyst_to_protobuf(analyst))

    g.userdata.update(obj.userdata)

    return g


def node_to_protobuf(obj: AbstractNode) -> graph_pb2.Node:
    if isinstance(obj, AtomNode):
        return atom_to_protobuf(obj)
    elif isinstance(obj, SchemeNode):
        return scheme_to_protobuf(obj)

    raise ValueError("Node type not supported")


def atom_to_protobuf(obj: AtomNode) -> graph_pb2.Node:
    """Export AtomNode object into PROTOBUF Node object."""
    proto = graph_pb2.Node(metadata=metadata_to_protobuf(obj.metadata))
    proto.userdata.update(obj.userdata)

    proto.atom.text = obj.plain_text

    if reference := obj.reference:
        proto.atom.reference.CopyFrom(reference_to_protobuf(reference))

    if participant := obj.participant:
        proto.atom.participant = participant.id

    return proto


def scheme_to_protobuf(obj: SchemeNode) -> graph_pb2.Node:
    """Export SchemeNode object into PROTOBUF Node object."""
    proto = graph_pb2.Node(metadata=metadata_to_protobuf(obj.metadata))
    proto.userdata.update(obj.userdata)

    if isinstance(obj.scheme, Support):
        proto.scheme.support = support2protobuf[obj.scheme]
    elif isinstance(obj.scheme, Attack):
        proto.scheme.attack = attack2protobuf[obj.scheme]
    elif isinstance(obj.scheme, Rephrase):
        proto.scheme.rephrase = rephrase2protobuf[obj.scheme]
    elif isinstance(obj.scheme, Preference):
        proto.scheme.preference = preference2protobuf[obj.scheme]

    proto.scheme.premise_descriptors.extend(obj.premise_descriptors)

    return proto


def analyst_to_protobuf(obj: Analyst) -> graph_pb2.Analyst:
    """Export Analyst object into a Graph's Analyst object in PROTOBUF format."""
    proto = graph_pb2.Analyst(
        name=obj.name or "",
        email=obj.email or "",
    )
    proto.userdata.update(obj.userdata)

    return proto


def edge_to_protobuf(obj: Edge) -> graph_pb2.Edge:
    """Export Edge object into PROTOBUF Edge format."""
    proto = graph_pb2.Edge(
        source=obj.source.id,
        target=obj.target.id,
        metadata=metadata_to_protobuf(obj.metadata),
    )
    proto.userdata.update(obj.userdata)

    return proto


def metadata_to_protobuf(obj: Metadata) -> graph_pb2.Metadata:
    proto = graph_pb2.Metadata()

    # if analyst := obj._analyst:
    #     proto.analyst = analyst.id

    dt.to_protobuf(obj.created, proto.created)
    dt.to_protobuf(obj.updated, proto.updated)

    return proto


def participant_to_protobuf(obj: Participant) -> graph_pb2.Participant:
    """Export Participant object into a Graph's Participant object in PROTOBUF format."""
    proto = graph_pb2.Participant(
        name=obj.name or "",
        username=obj.username or "",
        email=obj.email or "",
        url=obj.url or "",
        location=obj.location or "",
        description=obj.description or "",
        metadata=metadata_to_protobuf(obj.metadata),
    )
    proto.userdata.update(obj.userdata)

    return proto


def reference_to_protobuf(obj: Reference) -> graph_pb2.Reference:
    """Export Resource object into a Graph's Resource object in PROTOBUF format."""
    proto = graph_pb2.Reference(text=obj.plain_text)

    if resource := obj._resource:
        proto.resource = resource.id

    if offset := obj.offset:
        proto.offset = offset

    return proto


def resource_to_protobuf(obj: Resource) -> graph_pb2.Resource:
    """Export Resource object into a Graph's Resource object in PROTOBUF format."""
    proto = graph_pb2.Resource(
        text=obj.plain_text, metadata=metadata_to_protobuf(obj.metadata)
    )
    proto.userdata.update(obj.userdata)

    if title := obj.title:
        proto.title = title

    if source := obj.source:
        proto.source = source

    return proto



from __future__ import annotations

import typing as t


from arguebuf.models.userdata import Userdata
from arguebuf.services import utils


class Analyst:
    name: t.Optional[str]
    email: t.Optional[str]
    userdata: Userdata
    _id: str

    def __init__(
        self,
        name: t.Optional[str] = None,
        email: t.Optional[str] = None,
        userdata: t.Optional[Userdata] = None,
        id: t.Optional[str] = None,
    ) -> None:
        self.name = name
        self.email = email
        self.userdata = userdata or {}
        self._id = id or utils.uuid()

    @property
    def id(self) -> str:
        return self._id



from __future__ import absolute_import, annotations

import typing as t
from enum import Enum

from arg_services.graph.v1 import graph_pb2

from arguebuf.schemas import aif


class Support(Enum):
    """Enumeration of all available Argumentation Schemes (Walton et al.)

    .. autoclasssumm:: Support
        :autosummary-sections: Methods
    """

    DEFAULT = "Default"
    POSITION_TO_KNOW = "Position to Know"
    EXPERT_OPINION = "Expert Opinion"
    WITNESS_TESTIMONY = "Witness Testimony"
    POPULAR_OPINION = "Popular Opinion"
    POPULAR_PRACTICE = "Popular Practice"
    EXAMPLE = "Example"
    ANALOGY = "Analogy"
    PRACTICAL_REASONING_FROM_ANALOGY = "Practical Resoning from Analogy"
    COMPOSITION = "Composition"
    DIVISION = "Division"
    OPPOSITIONS = "Oppositions"
    RHETORICAL_OPPOSITIONS = "Rhetorical Oppositions"
    ALTERNATIVES = "Alternatives"
    VERBAL_CLASSIFICATION = "Verbal Classification"
    VERBAL_CLASSIFICATION_DEFINITION = "Definition to Verbal Classification"
    VERBAL_CLASSIFICATION_VAGUENESS = "Vagueness of a Verbal Classification"
    VERBAL_CLASSIFICATION_ARBITRARINESS = "Arbitrariness of a Verbal Classification"
    INTERACTION_OF_ACT_AND_PERSON = "Interaction of Act and Person"
    VALUES = "Values"
    POSITIVE_VALUES = "Positive Values"
    NEGATIVE_VALUES = "Negative Values"
    SACRIFICE = "Sacrifice"
    THE_GROUP_AND_ITS_MEMBERS = "The Group and its Members"
    PRACTICAL_REASONING = "Practical Reasoning"
    TWO_PERSON_PRACTICAL_REASONING = "Two-Person Practical Reasoning"
    WASTE = "Waste"
    SUNK_COSTS = "Sunk Costs"
    IGNORANCE = "Ignorance"
    EPISTEMIC_IGNORANCE = "Epistemic Ignorance"
    CAUSE_TO_EFFECT = "Cause to Effect"
    CORRELATION_TO_CAUSE = "Correlation to Cause"
    SIGN = "Sign"
    ABDUCTIVE = "Abductive"
    EVIDENCE_TO_HYPOTHESIS = "Evidence to Hypothesis"
    CONSEQUENCES = "Consequences"
    POSITIVE_CONSEQUENCES = "Positive Consequences"
    NEGATIVE_CONSEQUENCES = "Negative Consequences"
    PRAGMATIC_ALTERNATIVES = "Pragmatic Alternatives"
    THREAT = "Threat"
    FEAR_APPEAL = "Fear Appeal"
    DANGER_APPEAL = "Danger Appeal"
    NEED_FOR_HELP = "Need for Help"
    DISTRESS = "Distress"
    COMMITMENT = "Commitment"
    ETHOTIC = "Ethotic"
    GENERIC_AD_HOMINEM = "Generic ad Hominem"
    PRAGMATIC_INCONSISTENCY = "Pragmatic Inconsistency"
    INCONSISTENT_COMMITMENT = "Inconsistent Commitment"
    CIRCUMSTANTIAL_AD_HOMINEM = "Circumstantial Ad Hominem"
    BIAS = "Bias"
    BIAS_AD_HOMINEM = "Bias Ad Hominem"
    GRADUALISM = "Gradualism"
    SLIPPERY_SLOPE = "Slippery Slope"
    PRECEDENT_SLIPPERY_SLOPE = "Precedent Slippery Slope"
    SORITES_SLIPPERY_SLOPE = "Sorites Slippery Slope"
    VERBAL_SLIPPERY_SLOPE = "Verbal Slippery Slope"
    FULL_SLIPPERY_SLOPE = "Full Slippery Slope"
    CONSTITUTIVE_RULE_CLAIMS = "Constitutive Rule Claims"
    RULES = "Rules"
    EXCEPTIONAL_CASE = "Exceptional Case"
    PRECEDENT = "Precedent"
    PLEA_FOR_EXCUSE = "Plea for Excuse"
    PERCEPTION = "Perception"
    MEMORY = "Memory"
    # AUTHORITY = "Authority"
    # DILEMMA = "Dilemma"
    # MODUS_PONENS = "Modus Ponens"
    # DEFINITION = "Definition"


class Attack(Enum):
    """Enumeration of attacking schemes

    .. autoclasssumm:: Attack
        :autosummary-sections: Methods
    """

    DEFAULT = "Default"


class Preference(Enum):
    """Enumeration of preference schemes

    .. autoclasssumm:: Preference
        :autosummary-sections: Methods
    """

    DEFAULT = "Default"


class Rephrase(Enum):
    """Enumeration of rephrase schemes

    .. autoclasssumm:: Rephrase
        :autosummary-sections: Methods
    """

    DEFAULT = "Default"


Scheme = t.Union[Support, Attack, Rephrase, Preference]

support2protobuf = {
    item: graph_pb2.Support.Value(f"SUPPORT_{item.name}") for item in Support
}
protobuf2support = {value: key for key, value in support2protobuf.items()}

attack2protobuf = {
    item: graph_pb2.Attack.Value(f"ATTACK_{item.name}") for item in Attack
}
protobuf2attack = {value: key for key, value in attack2protobuf.items()}

rephrase2protobuf = {
    item: graph_pb2.Rephrase.Value(f"REPHRASE_{item.name}") for item in Rephrase
}
protobuf2rephrase = {value: key for key, value in rephrase2protobuf.items()}

preference2protobuf = {
    item: graph_pb2.Preference.Value(f"PREFERENCE_{item.name}") for item in Preference
}
protobuf2preference = {value: key for key, value in preference2protobuf.items()}

scheme2aif: t.Dict[t.Type[Scheme], aif.SchemeType] = {
    Support: "RA",
    Attack: "CA",
    Rephrase: "MA",
    Preference: "PA",
}
aif2scheme: t.Dict[aif.SchemeType, t.Optional[Scheme]] = {
    "RA": Support.DEFAULT,
    "CA": Attack.DEFAULT,
    "MA": Rephrase.DEFAULT,
    "PA": Preference.DEFAULT,
    "": None,
}

text2support: t.Dict[str, Support] = {
    "Alternatives": Support.ALTERNATIVES,
    "Analogy": Support.ANALOGY,
    "Arbitrary Verbal Classification": Support.VERBAL_CLASSIFICATION,
    "Argument From Authority": Support.DEFAULT,  # AUTHORITY
    "Argument From Goodwill": Support.DEFAULT,
    "Argument From Moral Virtue": Support.DEFAULT,
    "Argument From Practical Wisdom": Support.DEFAULT,
    "Argument From Virtue/Goodwill": Support.DEFAULT,
    "Argument From Wisdom/Goodwill": Support.DEFAULT,
    "Argument From Wisdom/Virtue": Support.DEFAULT,
    "Argument From Wisdom/Virtue/Goodwill": Support.DEFAULT,
    "Authority": Support.DEFAULT,  # AUTHORITY
    "Bias": Support.BIAS,
    "Causal Slippery Slope": Support.SLIPPERY_SLOPE,
    "Cause To Effect": Support.CAUSE_TO_EFFECT,
    "Circumstantial Ad Hominem": Support.CIRCUMSTANTIAL_AD_HOMINEM,
    "Commitment": Support.COMMITMENT,
    "Composition": Support.COMPOSITION,
    "Consequences": Support.CONSEQUENCES,
    "Correlation To Cause": Support.CORRELATION_TO_CAUSE,
    "Danger Appeal": Support.DANGER_APPEAL,
    "Default Inference": Support.DEFAULT,
    "Definitional": Support.DEFAULT,  # DEFINITION
    "Definition To Verbal Classification": Support.VERBAL_CLASSIFICATION,
    "Dilemma": Support.DEFAULT,  # DILEMMA
    "Direct Ad Hominem": Support.GENERIC_AD_HOMINEM,
    "Division": Support.DIVISION,
    "Efficient Cause": Support.DEFAULT,
    "Established Rule": Support.DEFAULT,
    "Ethotic": Support.ETHOTIC,
    "Evidence To Hypothesis": Support.EVIDENCE_TO_HYPOTHESIS,
    "Example": Support.EXAMPLE,
    "Exceptional Case": Support.EXCEPTIONAL_CASE,
    "Expert Opinion": Support.EXPERT_OPINION,
    "Falsification Of Hypothesis": Support.DEFAULT,
    "Fear Appeal": Support.FEAR_APPEAL,
    "Final Cause": Support.DEFAULT,
    "Formal Cause": Support.DEFAULT,
    "From-all-the-more-so-OR-all-the-less-so": Support.DEFAULT,
    "From-alternatives": Support.ALTERNATIVES,
    "From-analogy": Support.ANALOGY,
    "From-authority": Support.DEFAULT,  # AUTHORITY
    "From-conjugates-OR-derivates": Support.DEFAULT,
    "From-correlates": Support.DEFAULT,
    "From-definition": Support.DEFAULT,  # DEFINITION
    "From-description": Support.DEFAULT,
    "From-efficient-cause": Support.DEFAULT,
    "From-final-OR-instrumental-cause": Support.DEFAULT,
    "From-formal-cause": Support.DEFAULT,
    "From-genus-and-species": Support.DEFAULT,
    "From-material-cause": Support.DEFAULT,
    "From-ontological-implications": Support.DEFAULT,
    "From-opposition": Support.OPPOSITIONS,
    "From-parts-and-whole": Support.DIVISION,
    "From-place": Support.DEFAULT,
    "From-promising-and-warning": Support.DEFAULT,
    "From-termination-and-inception": Support.DEFAULT,
    "From-time": Support.DEFAULT,
    "Full Slippery Slope": Support.FULL_SLIPPERY_SLOPE,
    "Generic Ad Hominem": Support.GENERIC_AD_HOMINEM,
    "Gradualism": Support.GRADUALISM,
    "Ignorance": Support.IGNORANCE,
    "Inconsistent Commitment": Support.INCONSISTENT_COMMITMENT,
    "Informant Report": Support.DEFAULT,
    "Interaction Of Act And Person": Support.INTERACTION_OF_ACT_AND_PERSON,
    "Material Cause": Support.DEFAULT,
    "Mereological": Support.DEFAULT,
    "Modus Ponens": Support.DEFAULT,  # MODUS_PONENS
    "Need For Help": Support.NEED_FOR_HELP,
    "Negative Consequences": Support.NEGATIVE_CONSEQUENCES,
    "Opposition": Support.OPPOSITIONS,
    "Paraphrase": Support.DEFAULT,
    "Perception": Support.PERCEPTION,
    "Popular Opinion": Support.POPULAR_OPINION,
    "Popular Practice": Support.POPULAR_PRACTICE,
    "Position To Know": Support.POSITION_TO_KNOW,
    "Positive Consequences": Support.POSITIVE_CONSEQUENCES,
    "Practical Evaluation": Support.DEFAULT,
    "Practical Reasoning": Support.PRACTICAL_REASONING,
    "Practical Reasoning From Analogy": Support.PRACTICAL_REASONING_FROM_ANALOGY,
    "Pragmatic Argument From Alternatives": Support.PRAGMATIC_ALTERNATIVES,
    "Pragmatic Inconsistency": Support.PRAGMATIC_INCONSISTENCY,
    "Precedent Slippery Slope": Support.PRECEDENT_SLIPPERY_SLOPE,
    "Reframing": Support.DEFAULT,
    "Rules": Support.RULES,
    "Sign": Support.SIGN,
    "Two Person Practical Reasoning": Support.TWO_PERSON_PRACTICAL_REASONING,
    "Vagueness Of Verbal Classification": Support.VERBAL_CLASSIFICATION,
    "Vague Verbal Classification": Support.VERBAL_CLASSIFICATION,
    "Value Based Practical Reasoning": Support.PRACTICAL_REASONING,
    "Values": Support.VALUES,
    "Verbal Classification": Support.VERBAL_CLASSIFICATION,
    "Verbal Slippery Slope": Support.VERBAL_SLIPPERY_SLOPE,
    "Waste": Support.WASTE,
    "Witness Testimony": Support.WITNESS_TESTIMONY,
}

text2scheme: t.Dict[
    t.Type[Scheme],
    t.Union[
        t.Dict[str, Support],
        t.Dict[str, Attack],
        t.Dict[str, Rephrase],
        t.Dict[str, Preference],
    ],
] = {
    Support: text2support,
    Attack: {},
    Rephrase: {},
    Preference: {},
}



from __future__ import annotations

import typing as t

import pendulum

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.models.edge import Edge, warn_missing_nodes
from arguebuf.models.graph import Graph
from arguebuf.models.node import AbstractNode, AtomNode, SchemeNode
from arguebuf.models.scheme import aif2scheme, text2scheme
from arguebuf.schemas import aif
from arguebuf.services import dt, utils


def from_aif(
    obj: aif.Graph,
    name: t.Optional[str] = None,
    config: ConverterConfig = DefaultConverter,
) -> Graph:
    """Generate Graph structure from AIF argument graph file
    (reference: http://www.wi2.uni-trier.de/shared/publications/2019_LenzOllingerSahitajBergmann_ICCBR.pdf)

    """
    g = config.GraphClass(name)

    for aif_node in obj["nodes"]:
        node = (
            atom_from_aif(aif_node, config)
            if aif_node["type"] == "I"
            else scheme_from_aif(aif_node, config)
        )

        if node:
            g.add_node(node)

    for aif_edge in obj["edges"]:
        if edge := edge_from_aif(aif_edge, g.nodes, config):
            g.add_edge(edge)

    return g


def atom_from_aif(obj: aif.Node, config: ConverterConfig) -> AtomNode:
    """Generate AtomNode object from AIF Node object."""
    timestamp = dt.from_format(obj.get("timestamp"), aif.DATE_FORMAT) or pendulum.now()

    return config.AtomNodeClass(
        id=obj["nodeID"],
        metadata=config.MetadataClass(timestamp, timestamp),
        text=utils.parse(obj["text"], config.nlp),
    )


def scheme_from_aif(obj: aif.Node, config: ConverterConfig) -> t.Optional[SchemeNode]:
    """Generate SchemeNode object from AIF Node object."""

    aif_type = obj["type"]
    aif_scheme: str = obj.get("scheme", obj["text"])

    if aif_type in aif2scheme:
        scheme = aif2scheme[t.cast(aif.SchemeType, aif_type)]

        # TODO: Handle formatting like capitalization, spaces, underscores, etc.
        # TODO: Araucaria does not use spaces between scheme names
        # aif_scheme = re.sub("([A-Z])", r" \1", aif_scheme)
        if scheme and (found_scheme := text2scheme[type(scheme)].get(aif_scheme)):
            scheme = found_scheme

        timestamp = (
            dt.from_format(obj.get("timestamp"), aif.DATE_FORMAT) or pendulum.now()
        )

        return config.SchemeNodeClass(
            id=obj["nodeID"],
            metadata=config.MetadataClass(timestamp, timestamp),
            scheme=scheme,
        )

    return None


def edge_from_aif(
    obj: aif.Edge, nodes: t.Mapping[str, AbstractNode], config: ConverterConfig
) -> t.Optional[Edge]:
    """Generate Edge object from AIF Edge format."""
    source_id = obj.get("fromID")
    target_id = obj.get("toID")

    if source_id in nodes and target_id in nodes:
        return config.EdgeClass(
            id=obj["edgeID"],
            source=nodes[source_id],
            target=nodes[target_id],
        )
    else:
        warn_missing_nodes(obj["edgeID"], source_id, target_id)

    return None



import logging
import typing as t

import deepl
from multimethod import multimethod

import arguebuf as ag

log = logging.getLogger(__name__)


class Translator:
    translator: deepl.Translator
    source_lang: str
    target_lang: str

    def __init__(self, auth_key: str, source_lang: str, target_lang: str):
        self.source_lang = source_lang
        self.target_lang = target_lang
        self.translator = deepl.Translator(auth_key)

    def _deepl_translate(
        self, text: t.Union[str, t.Iterable[str]]
    ) -> t.Union[str, t.List[str]]:
        result = self.translator.translate_text(
            text,
            source_lang=self.source_lang,
            target_lang=self.target_lang,
            preserve_formatting=True,
        )
        if isinstance(result, list):
            return [entry.text for entry in result]
        else:
            return result.text

    @multimethod
    def translate(self):
        raise NotImplementedError()

    @translate.register
    def _(self, text: str) -> str:
        trans = self._deepl_translate(text)
        assert isinstance(trans, str)
        return trans

    @translate.register
    def _(self, texts: t.Iterable[str]) -> t.List[str]:
        trans = self._deepl_translate(texts)
        assert isinstance(trans, list)
        return trans

    @translate.register
    def _(self, graph: ag.Graph) -> None:
        original_resources = [
            resource.plain_text for resource in graph.resources.values()
        ]
        for resource, translation in zip(
            graph.resources.values(), self.translate(original_resources)
        ):
            resource.text = translation

        references = [
            atom.reference.plain_text
            for atom in graph.atom_nodes.values()
            if atom.reference is not None
        ]
        for atom, translation in zip(
            graph.atom_nodes.values(), self.translate(references)
        ):
            if atom.reference is not None:
                atom.reference.text = translation

        atoms = [atom.plain_text for atom in graph.atom_nodes.values()]
        for atom, translation in zip(graph.atom_nodes.values(), self.translate(atoms)):
            atom.text = translation

    @translate.register
    def _(self, graphs: t.Iterable[ag.Graph]) -> None:
        for graph in graphs:
            self.translate(graph)



import typing as t

import pendulum
from lxml import html

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.models.analyst import Analyst
from arguebuf.models.edge import Edge, warn_missing_nodes
from arguebuf.models.graph import Graph
from arguebuf.models.node import AbstractNode, AtomNode, SchemeNode
from arguebuf.models.resource import Resource
from arguebuf.models.scheme import aif2scheme, text2scheme
from arguebuf.schemas import aif, ova
from arguebuf.services import dt, utils


def from_ova(
    obj: ova.Graph,
    name: t.Optional[str] = None,
    config: ConverterConfig = DefaultConverter,
) -> Graph:
    """Generate Graph structure from OVA argument graph file (reference: http://ova.uni-trier.de/)."""
    g = config.GraphClass(name)

    analysis = obj["analysis"]
    resource = config.ResourceClass(
        utils.parse(analysis.get("plain_txt"), config.nlp),
        analysis.get("documentTitle"),
        analysis.get("documentSource"),
        dt.from_format(analysis.get("documentDate"), ova.DATE_FORMAT_ANALYSIS),
    )
    g.add_resource(resource)

    for participant in obj["participants"]:
        g.add_participant(
            config.ParticipantClass(
                name=f"{participant['firstname']} {participant['surname']}",
                id=str(participant["id"]),
            )
        )

    if analyst_name := obj["analysis"].get("annotatorName"):
        g.add_analyst(Analyst(name=analyst_name))

    for ova_node in obj["nodes"]:
        node = (
            atom_from_ova(ova_node, config)
            if ova_node.get("type") == "I"
            else scheme_from_ova(ova_node, config)
        )

        if node:
            g.add_node(node)

        if ova_node.get("major_claim") and isinstance(node, AtomNode):
            g._major_claim = node

    for ova_edge in obj["edges"]:
        if edge := edge_from_ova(ova_edge, g.nodes, config):
            g.add_edge(edge)

    if (analysis := obj.get("analysis")) and (raw_text := analysis.get("txt")):
        _inject_original_text(raw_text, g._atom_nodes, resource, config)

    return g


def _inject_original_text(
    raw_text: str,
    nodes: t.Mapping[str, AtomNode],
    resource: Resource,
    config: ConverterConfig,
) -> None:
    doc = html.fromstring(f"<html><head></head><body>{raw_text}</body></html>")
    text = ""
    body = doc.find("body")

    if body is not None:
        for elem in body.iter():
            # Span elements need special handling
            if elem.tag == "span":
                # The id is prefixed with 'node', e.g. 'node5'.
                node_key = str(elem.attrib["id"]).replace("node", "")
                if node := nodes.get(node_key):
                    node._reference = config.ReferenceClass(
                        resource, len(text), utils.parse(elem.text, config.nlp)
                    )

                if elem.text:
                    text += elem.text

            elif elem.tag == "br":
                text += "\n"

            elif elem.text:
                text += elem.text

            # Text after a tag should always be added to the overall text
            if elem.tail:
                text += elem.tail


def edge_from_ova(
    obj: ova.Edge, nodes: t.Mapping[str, AbstractNode], config: ConverterConfig
) -> t.Optional[Edge]:
    """Generate Edge object from OVA Edge format."""
    source_id = str(obj["from"]["id"])
    target_id = str(obj["to"]["id"])
    date = dt.from_format(obj.get("date"), ova.DATE_FORMAT) or pendulum.now()

    if source_id in nodes and target_id in nodes:
        return config.EdgeClass(
            id=utils.uuid(),
            source=nodes[source_id],
            target=nodes[target_id],
            metadata=config.MetadataClass(date, date),
        )
    else:
        warn_missing_nodes(None, source_id, target_id)

    return None


def scheme_from_ova(obj: ova.Node, config: ConverterConfig) -> t.Optional[SchemeNode]:
    """Generate SchemeNode object from OVA Node object."""

    ova_type = obj["type"]
    ova_scheme = obj["text"]

    if ova_type in aif2scheme:
        scheme = aif2scheme[t.cast(aif.SchemeType, ova_type)]

        if scheme and (found_scheme := text2scheme[type(scheme)].get(ova_scheme)):
            scheme = found_scheme

        premise_descriptors = [
            str(node_id)
            for description, node_id in obj["descriptors"].items()
            if not description.lower().startswith("s_conclusion")
        ]

        timestamp = dt.from_format(obj.get("date"), ova.DATE_FORMAT) or pendulum.now()

        return config.SchemeNodeClass(
            id=str(obj["id"]),
            metadata=config.MetadataClass(timestamp, timestamp),
            scheme=scheme,
            premise_descriptors=premise_descriptors,
        )

    return None


def atom_from_ova(obj: ova.Node, config: ConverterConfig) -> AtomNode:
    """Generate AtomNode object from OVA Node object."""
    timestamp = dt.from_format(obj.get("date"), ova.DATE_FORMAT) or pendulum.now()

    return config.AtomNodeClass(
        id=str(obj["id"]),
        metadata=config.MetadataClass(timestamp, timestamp),
        text=utils.parse(obj["text"], config.nlp),
    )



from __future__ import annotations

import typing as t
from dataclasses import dataclass
from enum import Enum

from arguebuf.models.analyst import Analyst
from arguebuf.models.edge import Edge
from arguebuf.models.graph import Graph
from arguebuf.models.metadata import Metadata
from arguebuf.models.node import AtomNode, SchemeNode
from arguebuf.models.participant import Participant
from arguebuf.models.reference import Reference
from arguebuf.models.resource import Resource
from arguebuf.models.typing import TextType


@dataclass
class ConverterConfig(t.Generic[TextType]):
    nlp: t.Optional[t.Callable[[str], TextType]] = None
    GraphClass: t.Type[Graph] = Graph
    AtomNodeClass: t.Type[AtomNode] = AtomNode
    SchemeNodeClass: t.Type[SchemeNode] = SchemeNode
    EdgeClass: t.Type[Edge] = Edge
    AnalystClass: t.Type[Analyst] = Analyst
    MetadataClass: t.Type[Metadata] = Metadata
    ParticipantClass: t.Type[Participant] = Participant
    ReferenceClass: t.Type[Reference] = Reference
    ResourceClass: t.Type[Resource] = Resource


DefaultConverter = ConverterConfig[str]()


class GraphFormat(str, Enum):
    ARGUEBUF = "arguebuf"
    AIF = "aif"



import typing as t
from enum import Enum

from graphviz import Digraph

GraphvizGraph = t.Union[Digraph, t.Any]


class EdgeStyle(Enum):
    BEZIER = "curved"
    STRAIGHT = "line"
    STEP = "ortho"



from __future__ import annotations

import re
import typing as t
from collections.abc import Iterable
from dataclasses import asdict, dataclass, field, fields
from pathlib import Path

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.converters.from_path import from_file, from_folder
from arguebuf.models.graph import Graph

format2glob = {
    "aif": "*.json",
    "aml": "*.aml",
    "argdown": "*.json",
    "arggraph": "*.xml",
    "arguebuf": "*.json",
    "brat": "*.ann",
    "kialo": "*.txt",
    "microtexts": "*.xml",
    "ova": "*.json",
    "protobuf": "*.json",
    None: "*",
}

DEFAULT_GLOB = "*"


@dataclass
class FilesystemFilter:
    name: str
    kwargs: t.Dict[str, str]

    def __init__(self, name: str, **kwargs: str) -> None:
        self.name = name
        self.kwargs = kwargs

    @property
    def glob(self) -> str:
        return format2glob.get(self.kwargs.get("format"), DEFAULT_GLOB)

    @classmethod
    def from_path(cls, path: Path) -> FilesystemFilter:
        kwargs: dict[str, str] = {}
        entries = path.name.replace(" ", "").split(",")

        for entry in entries:
            key, *values = entry.split("=")
            value = values[0] if len(values) == 1 else "true"
            kwargs[key] = value

        return cls(path.parent.name, **kwargs)


class CasebaseFilter:
    name: re.Pattern
    cases: t.Optional[re.Pattern]
    kwargs: t.Dict[str, re.Pattern]

    def __init__(self, name: str, cases: t.Optional[str] = None, **kwargs: str):
        self.name = re.compile(name)
        self.cases = re.compile(cases) if cases is not None else None
        self.kwargs = {}

        for key, value in kwargs.items():
            self.kwargs[key] = re.compile(value)

    def __eq__(self, other: FilesystemFilter) -> bool:
        return self.kwargs == other.kwargs and self >= other

    def __ge__(self, other: FilesystemFilter) -> bool:
        return all(
            pattern.match(other.kwargs.get(kwarg, "null"))
            for kwarg, pattern in self.kwargs.items()
        )


def from_casebase(
    basepath: t.Union[Path, str],
    include: t.Union[CasebaseFilter, t.Iterable[CasebaseFilter]],
    exclude: t.Union[CasebaseFilter, t.Iterable[CasebaseFilter], None] = None,
    glob: str = "*/*",
    config: ConverterConfig = DefaultConverter,
    strict_equal: bool = False,
):
    graphs: dict[Path, Graph] = {}

    if not isinstance(basepath, Path):
        basepath = Path(basepath)

    if not isinstance(include, Iterable):
        include = [include]

    # if isinstance(exclude, CasebaseFilter):
    #     exclude = [exclude]

    for filter in include:
        for path in sorted(basepath.glob(glob)):
            if (
                path.is_dir()
                and filter.name.match(path.parent.name)
                and not path.parent.name.startswith(".")
                and not path.name.startswith(".")
            ):
                graphs.update(_from_casebase_single(filter, path, config, strict_equal))

    return graphs


def _from_casebase_single(
    user_filter: CasebaseFilter,
    path: Path,
    config: ConverterConfig,
    strict_equal: bool,
):
    filesystem_filter = FilesystemFilter.from_path(path)

    if (strict_equal and user_filter == filesystem_filter) or (
        not strict_equal and user_filter >= filesystem_filter
    ):
        glob = f"**/{filesystem_filter.glob}"

        if user_filter.cases is not None:
            graphs = {}

            for file in sorted(path.glob(glob)):
                if file.is_file() and user_filter.cases.match(
                    str(file.relative_to(path))
                ):
                    graphs[file] = from_file(file, config=config)

            return graphs

        return from_folder(path, glob, config=config)

    return {}



from __future__ import absolute_import, annotations

import logging
import typing as t


from arguebuf.models.metadata import Metadata
from arguebuf.models.node import AbstractNode
from arguebuf.models.userdata import Userdata
from arguebuf.services import utils

log = logging.getLogger(__name__)


def warn_missing_nodes(
    edge_id: t.Optional[str], source_id: t.Optional[str], target_id: t.Optional[str]
) -> None:
    log.warning(
        f"Skipping edge '{edge_id}': Source '{source_id}' or target '{target_id}' not found."
    )


class Edge:
    """Edge in AIF format. Connection from one Node object to another Node object."""

    __slots__ = (
        "_id",
        "_source",
        "_target",
        "metadata",
        "userdata",
    )

    _id: str
    _source: AbstractNode
    _target: AbstractNode
    metadata: Metadata
    userdata: Userdata

    def __init__(
        self,
        source: AbstractNode,
        target: AbstractNode,
        metadata: t.Optional[Metadata] = None,
        userdata: t.Optional[Userdata] = None,
        id: t.Optional[str] = None,
    ):
        # if isinstance(source, AtomNode) and isinstance(target, AtomNode):
        #     raise ValueError("Cannot create an edge between two atom nodes.")

        self._id = id or utils.uuid()
        self._source = source
        self._target = target
        self.metadata = metadata or Metadata()
        self.userdata = userdata or {}

        self.__post_init__()

    def __post_init__(self):
        pass

    def __repr__(self):
        return utils.class_repr(
            self,
            [str(self._id), f"{self._source.__repr__()}->{self._target.__repr__()}"],
        )

    @property
    def id(self) -> str:
        return self._id

    @property
    def source(self) -> AbstractNode:
        """Gives the 'From'-Node."""
        return self._source

    @property
    def target(self) -> AbstractNode:
        """Gives the 'To'-Node."""
        return self._target



from typer import Typer

from . import graph, text

cli = Typer()
cli.add_typer(
    graph.cli, name="graph", help="Commands for dealing with argument graphs."
)
cli.add_typer(text.cli, name="text", help="Commands for dealing with plain texts.")

if __name__ == "__main__":
    cli()



import typing as t

from arguebuf.schemas.aif import NodeType

DATE_FORMAT = "DD/MM/YYYY - HH:mm:ss"
DATE_FORMAT_ANALYSIS = "DD/MM/YYYY"

Node = t.TypedDict(
    "Node",
    {
        "id": int,
        "x": float,
        "y": float,
        "color": str,
        "text": str,
        "type": NodeType,
        "scheme": str,
        "descriptors": t.Dict[str, int],
        "cqdesc": t.Dict[str, t.Any],
        "visible": bool,
        "participantID": str,
        "w": float,
        "h": float,
        #
        # Specific to original OVA
        #
        "imgurl": t.Optional[str],
        #
        # Specific to ReCAP OVA
        #
        "majorClaim": t.Optional[bool],
        "is_check_worthy": t.Optional[str],
        "source": t.Optional[str],
        "text_begin": t.Optional[t.List[int]],
        "text_end": t.Optional[t.List[int]],
        "text_length": t.Optional[t.List[int]],
        "comment": t.Optional[str],
        "annotator": t.Optional[str],
        "date": t.Optional[str],
    },
)

Edge = t.TypedDict(
    "Edge",
    {
        "from": Node,
        "to": Node,
        "visible": bool,
        #
        # Specific to ReCAP OVA
        #
        "annotator": t.Optional[str],
        "date": t.Optional[str],
    },
)

Participant = t.TypedDict("Participant", {"id": int, "firstname": str, "surname": str})

Analysis = t.TypedDict(
    "Analysis",
    {
        "txt": str,
        #
        # Specific to ReCAP OVA
        #
        "plain_txt": t.Optional[str],
        "annotatorName": t.Optional[str],
        "documentSource": t.Optional[str],
        "documentTitle": t.Optional[str],
        "ovaVersion": t.Optional[str],
    },
)

Graph = t.TypedDict(
    "Graph",
    {
        "nodes": t.List[Node],
        "edges": t.List[Edge],
        "participants": t.List[Participant],
        "analysis": Analysis,
    },
)



from __future__ import annotations

import typing as t
from pathlib import Path

from arguebuf.converters.config import GraphFormat
from arguebuf.converters.to_io import to_io
from arguebuf.models.graph import Graph


def to_file(
    graph: Graph,
    path: t.Union[Path, str],
    format: GraphFormat = GraphFormat.ARGUEBUF,
    pretty: bool = False,
) -> None:
    """Export strucure of Graph instance into structure of File/Folder format."""
    if isinstance(path, str):
        path = Path(path)

    if path.is_dir() or not path.suffix:
        path = path / f"{graph.name}.json"

    with path.open("w", encoding="utf-8") as file:
        to_io(graph, file, format, pretty)


to_folder = to_file



from __future__ import annotations

import typing as t

from arg_services.graph.v1 import graph_pb2

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.models.edge import Edge, warn_missing_nodes
from arguebuf.models.graph import Graph
from arguebuf.models.metadata import Metadata
from arguebuf.models.node import AbstractNode, AtomNode, SchemeNode
from arguebuf.models.participant import Participant
from arguebuf.models.reference import Reference
from arguebuf.models.resource import Resource
from arguebuf.models.scheme import (
    protobuf2attack,
    protobuf2preference,
    protobuf2rephrase,
    protobuf2support,
)
from arguebuf.services import dt, utils


def from_protobuf(
    obj: graph_pb2.Graph,
    name: t.Optional[str] = None,
    config: ConverterConfig = DefaultConverter,
) -> Graph:
    """Generate Graph structure from PROTOBUF argument graph file.(Link?)"""
    g = config.GraphClass(name)

    for resource_id, resource in obj.resources.items():
        g.add_resource(
            config.ResourceClass(
                utils.parse(resource.text, config.nlp),
                resource.title,
                resource.source,
                dt.from_protobuf(resource.timestamp),
                metadata_from_protobuf(resource.metadata, config),
                dict(resource.userdata.items()),
                resource_id,
            )
        )

    for participant_id, participant in obj.participants.items():
        g.add_participant(
            config.ParticipantClass(
                participant.name,
                participant.username,
                participant.email,
                participant.url,
                participant.location,
                participant.description,
                metadata_from_protobuf(participant.metadata, config),
                dict(participant.userdata.items()),
                participant_id,
            )
        )

    for analyst_id, analyst in obj.analysts.items():
        g.add_analyst(
            config.AnalystClass(
                analyst.name,
                analyst.email,
                dict(analyst.userdata.items()),
                analyst_id,
            )
        )

    for node_id, node in obj.nodes.items():
        if node.WhichOneof("type") == "atom":
            g.add_node(
                atom_from_protobuf(node_id, node, g.resources, g.participants, config)
            )
        elif node.WhichOneof("type") == "scheme":
            g.add_node(scheme_from_protobuf(node_id, node, config))
        # TODO: Raise error if node is neither scheme nor atom

    for edge_id, proto_edge in obj.edges.items():
        if edge := edge_from_protobuf(edge_id, proto_edge, g.nodes, config):
            g.add_edge(edge)

    major_claim = g.nodes[obj.major_claim] if obj.major_claim else None

    if major_claim and isinstance(major_claim, AtomNode):
        g._major_claim = major_claim

    g.userdata.update(obj.userdata)
    g.metadata = metadata_from_protobuf(obj.metadata, config)
    g.library_version = obj.library_version
    g.schema_version = obj.schema_version

    return g


def atom_from_protobuf(
    id: str,
    obj: graph_pb2.Node,
    resources: t.Mapping[str, Resource],
    participants: t.Mapping[str, Participant],
    config: ConverterConfig,
) -> AtomNode:
    """Generate AtomNode object from PROTOBUF Node object."""
    return config.AtomNodeClass(
        utils.parse(obj.atom.text, config.nlp),
        reference_from_protobuf(obj.atom.reference, resources, config),
        participants.get(obj.atom.participant),
        metadata_from_protobuf(obj.metadata, config),
        dict(obj.userdata.items()),
        id=id,
    )


def scheme_from_protobuf(
    id: str, obj: graph_pb2.Node, config: ConverterConfig
) -> SchemeNode:
    """Generate SchemeNode object from OVA Node object."""

    scheme_type = obj.scheme.WhichOneof("type")
    scheme = None

    if scheme_type is None:
        scheme = None
    elif scheme_type == "support":
        scheme = protobuf2support[obj.scheme.support]
    elif scheme_type == "attack":
        scheme = protobuf2attack[obj.scheme.attack]
    elif scheme_type == "rephrase":
        scheme = protobuf2rephrase[obj.scheme.rephrase]
    elif scheme_type == "preference":
        scheme = protobuf2preference[obj.scheme.preference]

    return config.SchemeNodeClass(
        scheme,
        list(obj.scheme.premise_descriptors),
        metadata_from_protobuf(obj.metadata, config),
        dict(obj.userdata.items()),
        id=id,
    )


def edge_from_protobuf(
    id: str,
    obj: graph_pb2.Edge,
    nodes: t.Mapping[str, AbstractNode],
    config: ConverterConfig,
) -> t.Optional[Edge]:
    """Generate Edge object from PROTOBUF Edge format."""
    if obj.source in nodes and obj.target in nodes:
        return config.EdgeClass(
            nodes[obj.source],
            nodes[obj.target],
            metadata_from_protobuf(obj.metadata, config),
            dict(obj.userdata.items()),
            id=id,
        )
    else:
        warn_missing_nodes(id, obj.source, obj.target)

    return None


def metadata_from_protobuf(
    obj: graph_pb2.Metadata, config: ConverterConfig
) -> Metadata:
    return config.MetadataClass(
        dt.from_protobuf(obj.created),
        dt.from_protobuf(obj.updated),
        # analysts[obj.analyst]
    )


def reference_from_protobuf(
    obj: graph_pb2.Reference,
    resources: t.Mapping[str, Resource],
    config: ConverterConfig,
) -> t.Optional[Reference]:
    """Generate Resource object from PROTOBUF format Graph's Resource object."""
    if obj.text:
        if obj.resource:
            return config.ReferenceClass(
                resources[obj.resource],
                obj.offset,
                utils.parse(obj.text, config.nlp),
            )

        else:
            return config.ReferenceClass(
                None,
                None,
                utils.parse(obj.text, config.nlp),
            )

    return None



import typing as t

import networkx as nx

from arguebuf.models.edge import Edge
from arguebuf.models.graph import Graph
from arguebuf.models.node import AbstractNode, AtomNode, SchemeNode


def to_networkx(
    graph: Graph,
    graph_attrs: t.Optional[t.MutableMapping[str, t.Callable[[Graph], t.Any]]] = None,
    atom_attrs: t.Optional[t.MutableMapping[str, t.Callable[[AtomNode], t.Any]]] = None,
    scheme_attrs: t.Optional[
        t.MutableMapping[str, t.Callable[[SchemeNode], t.Any]]
    ] = None,
    edge_attrs: t.Optional[t.MutableMapping[str, t.Callable[[Edge], t.Any]]] = None,
) -> nx.DiGraph:
    """Transform the argument graph for use with the library `NetworkX`

    This library allows you to apply advanced graph-related algorithms directly on your argument graphs.
    For instance, shortest paths and various distances can be computed.
    [Documentation](https://networkx.org/documentation/stable/reference/index.html)

    It is possible to add arbitrary attributes to the resulting graph and its elements.
    For this, you need to pass a dictionary with the desired name of the attribute and a function that is used to compute the attribute's value.
    The function will be passed the corresponding element as its only parameter.
    For instance, you could pass `atom_attrs={"text": lambda node: node.plain_text}` to set a `text` attribute for atom nodes.

    Args:
        graph_attrs: Attribute functions for the whole graph.
        atom_attrs: Attribute functions for the atom nodes.
        scheme_attrs: Attribute functions for the scheme nodes.
        edge_attrs: Attribute functions for the edges.

    Returns:
        Instance of the *directed* `NetworkX` graph.

    Examples:
        >>> g = Graph("Test")
        >>> n1 = AtomNode("Node1")
        >>> n2 = AtomNode("Node2")
        >>> e = Edge(n1, n2)
        >>> g.add_edge(e)
        >>> gnx = to_networkx(g)
        >>> gnx.number_of_nodes()
        2
    """

    if graph_attrs is None:
        graph_attrs = {}

    g = nx.DiGraph(None, **{key: func(graph) for key, func in graph_attrs.items()})

    for node in graph.atom_nodes.values():
        node_to_nx(node, g, atom_attrs)  # type: ignore

    for node in graph.scheme_nodes.values():
        node_to_nx(node, g, scheme_attrs)  # type: ignore

    for edge in graph.edges.values():
        edge_to_nx(edge, g, edge_attrs)

    return g


def edge_to_nx(
    edge: Edge,
    g: nx.DiGraph,
    attrs: t.Optional[t.MutableMapping[str, t.Callable[[Edge], t.Any]]] = None,
) -> None:
    """Submethod used to export Graph object g into NX Graph format."""

    if attrs is None:
        attrs = {}

    g.add_edge(
        edge.source.id,
        edge.target.id,
        **{key: func(edge) for key, func in attrs.items()},
    )


def node_to_nx(
    node: AbstractNode,
    g: nx.DiGraph,
    attrs: t.Optional[t.MutableMapping[str, t.Callable[[AbstractNode], t.Any]]] = None,
) -> None:
    """Submethod used to export Graph object g into NX Graph format."""
    if attrs is None:
        attrs = {}

    if "label" not in attrs:
        attrs["label"] = lambda x: x.label

    g.add_node(node._id, **{key: func(node) for key, func in attrs.items()})



from __future__ import annotations


from arguebuf.models.edge import Edge
from arguebuf.models.graph import Graph
from arguebuf.models.node import NO_SCHEME_LABEL, AbstractNode, AtomNode, SchemeNode
from arguebuf.models.scheme import scheme2aif
from arguebuf.schemas import aif
from arguebuf.services import dt


def to_aif(obj: Graph) -> aif.Graph:
    """Export structure of Graph instance to AIF argument graph format."""
    return {
        "nodes": [node_to_aif(node) for node in obj.nodes.values()],
        "edges": [edge_to_aif(edge) for edge in obj.edges.values()],
        "locutions": [],
    }


def node_to_aif(obj: AbstractNode) -> aif.Node:
    if isinstance(obj, AtomNode):
        return atom_to_aif(obj)
    elif isinstance(obj, SchemeNode):
        return scheme_to_aif(obj)

    raise ValueError("Node type not supported.")


def scheme_to_aif(obj: SchemeNode) -> aif.Node:
    """Export SchemeNode object into AIF Node object."""

    return {
        "nodeID": obj.id,
        "timestamp": dt.to_format(obj.metadata.updated, aif.DATE_FORMAT),
        "text": obj.scheme.value if obj.scheme else NO_SCHEME_LABEL,
        "type": scheme2aif[type(obj.scheme)] if obj.scheme else "",
    }


def atom_to_aif(obj: AtomNode) -> aif.Node:
    """Export AtomNode object into AIF Node object."""
    return {
        "nodeID": obj.id,
        "timestamp": dt.to_format(obj.metadata.updated, aif.DATE_FORMAT),
        "text": obj.plain_text,
        "type": "I",
    }


def edge_to_aif(obj: Edge) -> aif.Edge:
    """Export Edge object into AIF Edge format."""
    return {
        "edgeID": str(obj.id),
        "fromID": str(obj.source.id),
        "toID": str(obj.target.id),
        "formEdgeID": None,
    }



from __future__ import annotations

import json
import typing as t

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.converters.from_dict import from_dict
from arguebuf.models.graph import Graph


def from_json(
    obj: t.TextIO,
    name: t.Optional[str] = None,
    config: ConverterConfig = DefaultConverter,
) -> Graph:
    """Generate Graph structure from JSON argument graph file(Link?)."""
    return from_dict(json.load(obj), name, config)



import typing as t


class Node(t.TypedDict):
    id: str
    title: str  # title of argument/statement
    type: str  # check if "argument-map-node", or "statment-map-node"
    labelTitle: str  # title of argument/statement
    labelText: str  # text of argument/statement


class Edge(t.TypedDict):
    id: str
    type: str
    relationType: str
    source: str
    target: str


class Map(t.TypedDict):
    nodes: t.List[Node]
    edges: t.List[Edge]


class Graph(t.TypedDict):
    map: Map



import typing as t

import pendulum

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.models.edge import Edge, warn_missing_nodes
from arguebuf.models.graph import Graph
from arguebuf.models.metadata import Metadata
from arguebuf.models.node import AbstractNode, Attack, Support
from arguebuf.schemas import argdown


def from_argdown(
    obj: argdown.Graph,
    name: t.Optional[str] = None,
    config: ConverterConfig = DefaultConverter,
) -> Graph:
    """Generate Graph structure from JSON Argdown argument graph file
    (reference: https://argdown.org/).
    """
    g = config.GraphClass(name)

    timestamp = pendulum.now()

    # Every node in obj["nodes"] is a atom node
    for argdown_node in obj["map"]["nodes"]:
        node = config.AtomNodeClass.from_argdown_json(argdown_node, config.nlp)

        if node:
            g.add_node(node)

    for argdown_edge in obj["map"]["edges"]:
        if edge := edge_from_argdown(argdown_edge, g.nodes, config.EdgeClass):
            g.add_edge(edge)
            if argdown_edge["relationType"] == "support":
                scheme = Support.DEFAULT
            elif argdown_edge["relationType"] == "attack":
                scheme = Attack.DEFAULT
            elif argdown_edge["relationType"] == "contradictory":
                scheme = Attack.DEFAULT
            elif argdown_edge["relationType"] == "undercut":
                scheme = Support.DEFAULT
            else:
                scheme = Support.DEFAULT
            # create scheme_node for edge
            scheme_node = config.SchemeNodeClass(
                metadata=Metadata(timestamp, timestamp),
                scheme=scheme,
            )
            g.add_node(scheme_node)

            # create edge from source to scheme_node and an edge from scheme_node to target
            g.add_edge(Edge(edge.source, scheme_node))
            g.add_edge(Edge(scheme_node, edge.target))

    metadata = Metadata(timestamp, timestamp)
    g.metadata = metadata

    return g


def edge_from_argdown(
    obj: argdown.Edge,
    nodes: t.Mapping[str, AbstractNode],
    edge_class: t.Type[Edge] = Edge,
) -> t.Optional[Edge]:
    """Generate Edge object from Argdown JSON Edge format."""
    if "from" in obj:
        source_id = obj["from"]  # type: ignore
    else:
        source_id = obj["source"]

    if "to" in obj:
        target_id = obj["to"]  # type: ignore
    else:
        target_id = obj["target"]

    edge_id = obj["id"]

    if source_id in nodes and target_id in nodes:
        return edge_class(
            id=edge_id,
            source=nodes[source_id],
            target=nodes[target_id],
        )
    else:
        warn_missing_nodes(edge_id, source_id, target_id)

    return None



from .analyst import Analyst
from .edge import Edge
from .graph import Graph
from .metadata import Metadata
from .node import AbstractNode, AtomNode, AtomOrSchemeNode, SchemeNode
from .participant import Participant
from .reference import Reference
from .resource import Resource
from .scheme import Attack, Preference, Rephrase, Scheme, Support
from .userdata import Userdata

__all__ = (
    "Analyst",
    "Edge",
    "Graph",
    "Metadata",
    "AbstractNode",
    "AtomNode",
    "AtomOrSchemeNode",
    "SchemeNode",
    "Participant",
    "Reference",
    "Resource",
    "Attack",
    "Preference",
    "Rephrase",
    "Scheme",
    "Support",
    "Userdata",
)



import shutil
import typing as t
from pathlib import Path

import typer
from arguebuf.cli.translator import Translator

from . import model

cli = typer.Typer()


@cli.command()
def translate(
    input_folder: Path,
    source_lang: str,
    target_lang: str,
    auth_key: str,
    input_glob: str,
    output_suffix: str,
    output_folder: t.Optional[Path] = None,
    clean: bool = False,
    overwrite: bool = False,
    start: int = 1,
) -> None:
    if not output_folder:
        output_folder = input_folder

    if clean:
        shutil.rmtree(output_folder)
        output_folder.mkdir()

    paths = model.PathPair.create(
        input_folder, output_folder, input_glob, output_suffix
    )
    translator = Translator(auth_key, source_lang, target_lang)
    bar: t.Iterable[model.PathPair]

    with typer.progressbar(
        paths[start - 1 :],
        item_show_func=model.PathPair.label,
        show_pos=True,
    ) as bar:
        for path_pair in bar:
            if overwrite or not path_pair.target.exists():
                with path_pair.source.open("r") as file:
                    source_text = file.read()

                translation = translator.translate(source_text)

                with path_pair.target.open("w") as file:
                    file.write(translation)



import typing as t

import pendulum
from google.protobuf import timestamp_pb2
from pendulum.datetime import DateTime


def from_format(text: t.Optional[str], format: str) -> t.Optional[DateTime]:
    return pendulum.from_format(text, format) if text else None


def to_format(dt: t.Optional[DateTime], format: str) -> str:
    return dt.format(format) if dt else ""


def from_protobuf(dt: timestamp_pb2.Timestamp) -> DateTime:
    return pendulum.instance(dt.ToDatetime()) if dt else pendulum.now()


def to_protobuf(dt: t.Optional[DateTime], obj: timestamp_pb2.Timestamp) -> None:
    if dt:
        obj.FromDatetime(dt)



from .config import ConverterConfig, GraphFormat
from .copy import copy
from .from_aif import from_aif
from .from_aml import from_aml
from .from_argdown import from_argdown
from .from_brat import from_brat
from .from_casebase import CasebaseFilter, from_casebase
from .from_dict import from_dict
from .from_io import from_io
from .from_json import from_json
from .from_kialo import from_kialo
from .from_microtexts import from_microtexts
from .from_ova import from_ova
from .from_path import from_file, from_folder
from .from_protobuf import from_protobuf
from .from_sadface import from_sadface
from .to_aif import to_aif
from .to_dict import to_dict
from .to_graphviz import render, to_graphviz
from .to_io import to_io
from .to_json import to_json
from .to_networkx import to_networkx
from .to_path import to_file, to_folder
from .to_protobuf import to_protobuf

__all__ = (
    "ConverterConfig",
    "GraphFormat",
    "copy",
    "from_aif",
    "from_aml",
    "from_argdown",
    "from_brat",
    "CasebaseFilter",
    "from_casebase",
    "from_dict",
    "from_io",
    "from_json",
    "from_kialo",
    "from_microtexts",
    "from_ova",
    "from_file",
    "from_folder",
    "from_protobuf",
    "from_sadface",
    "to_aif",
    "to_dict",
    "render",
    "to_graphviz",
    "to_io",
    "to_json",
    "to_networkx",
    "to_file",
    "to_folder",
    "to_protobuf",
)



import typing as t

NodeType = t.Literal["scheme", "atom"]
DATE_FORMAT = "YYYY-MM-DDTHH:mm:ss"


class Node(t.TypedDict):
    id: str
    text: str
    type: NodeType
    name: str
    sources: t.List[str]
    metadata: t.Dict[str, t.Any]


class Edge(t.TypedDict):
    id: str
    source_id: str
    target_id: str


class Core(t.TypedDict):
    analyst_email: str
    analyst_name: str
    created: str
    description: str
    edited: str
    id: str
    notes: str
    title: str
    version: str


class Metadata(t.TypedDict):
    core: Core


class Graph(t.TypedDict):
    nodes: t.List[Node]
    edges: t.List[Edge]
    metadata: Metadata



from __future__ import absolute_import, annotations

import typing as t
from collections import abc
from uuid import uuid1


def uuid() -> str:
    return str(uuid1())


def _class_name(obj) -> str:
    return obj.__class__.__name__


def class_repr(obj, attributes: t.Iterable[str]) -> str:
    return f"{_class_name(obj)}({', '.join(attributes)})"


def xstr(data: t.Any) -> str:
    return "" if data is None else str(data)


def parse(text: t.Optional[str], nlp: t.Optional[t.Callable[[str], t.Any]]) -> t.Any:
    if nlp:
        if text is None:
            return nlp("")

        try:
            out = nlp(text)
        except ValueError:
            out = nlp("")

        return out

    return text


def type_error(actual: t.Type, expected: t.Type) -> str:
    return f"Expected type '{expected}', but got '{actual}'. Make sure that you are passing the correct method arguments."


def duplicate_key_error(name: str, key: str) -> str:
    return f"Graph '{name}' already contains an element with key '{key}'. The keys have to be unique within each graph."


def missing_key_error(name: str, key: str) -> str:
    return f"Graph '{name}' does not contain an element with key '{key}'. It cannot be removed."


T = t.TypeVar("T")
U = t.TypeVar("U")


class ImmutableList(abc.Sequence[T]):
    """Read-only view."""

    __slots__ = "_store"

    _store: t.MutableSequence[T]

    def __init__(self, items: t.Optional[t.MutableSequence[T]] = None):
        self._store = items or list()

    def __len__(self) -> int:
        return self._store.__len__()

    @t.overload
    def __getitem__(self, key: int) -> T:
        pass  # Don't put code here

    @t.overload
    def __getitem__(self, key: slice) -> t.Sequence[T]:
        pass  # Don't put code here

    def __getitem__(self, key: t.Union[int, slice]) -> t.Union[T, t.Sequence[T]]:
        return self._store.__getitem__(key)

    def __repr__(self) -> str:
        return self._store.__repr__()

    def __str__(self) -> str:
        return self._store.__str__()


class ImmutableSet(abc.Set[T]):
    """Read-only view."""

    __slots__ = "_store"

    _store: t.MutableSet[T]

    def __init__(self, items: t.Optional[t.MutableSet[T]] = None):
        self._store = items or set()

    def __len__(self) -> int:
        return self._store.__len__()

    def __contains__(self, item: object) -> bool:
        return self._store.__contains__(item)

    def __iter__(self) -> t.Iterator[T]:
        return self._store.__iter__()

    def __repr__(self) -> str:
        return self._store.__repr__()

    def __str__(self) -> str:
        return self._store.__str__()


class ImmutableDict(t.Mapping[T, U]):
    """Read-only view."""

    __slots__ = "_store"

    _store: t.MutableMapping[T, U]

    def __init__(self, items: t.Optional[t.MutableMapping[T, U]] = None):
        self._store = items or dict()

    def __len__(self) -> int:
        return self._store.__len__()

    def __getitem__(self, key: T) -> U:
        return self._store.__getitem__(key)

    def __contains__(self, key: T) -> bool:
        return self._store.__contains__(key)

    def __iter__(self) -> t.Iterator:
        return self._store.__iter__()

    def __repr__(self) -> str:
        return self._store.__repr__()

    def __str__(self) -> str:
        return self._store.__str__()



from __future__ import annotations

import typing as t

import pendulum

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.models.edge import Edge, warn_missing_nodes
from arguebuf.models.graph import Graph
from arguebuf.models.node import (
    AbstractNode,
    AtomNode,
    Attack,
    Preference,
    Rephrase,
    SchemeNode,
    Support,
)
from arguebuf.schemas import sadface
from arguebuf.services import dt, utils


def from_sadface(
    obj: sadface.Graph,
    name: t.Optional[str] = None,
    config: ConverterConfig = DefaultConverter,
) -> Graph:
    """Generate Graph structure from SADFace argument graph file
    (reference: https://github.com/Open-Argumentation/SADFace/blob/master/examples/hangback/data.json).
    """
    g = config.GraphClass(name)

    for sadface_node in obj["nodes"]:
        node = (
            atom_from_sadface(sadface_node, config)
            if sadface_node["type"] == "atom"
            else scheme_from_sadface(sadface_node, config)
        )

        if node:
            g.add_node(node)

    for sadface_edge in obj["edges"]:
        if edge := edge_from_sadface(sadface_edge, g.nodes, config):
            g.add_edge(edge)

    # create
    # object
    created = dt.from_format(obj["metadata"]["core"]["created"], sadface.DATE_FORMAT)
    updated = dt.from_format(obj["metadata"]["core"]["edited"], sadface.DATE_FORMAT)
    metadata = config.MetadataClass(created, updated)
    g.metadata = metadata

    # create Analyst object
    analyst = config.AnalystClass(
        name=obj["metadata"]["core"]["analyst_name"],
        email=obj["metadata"]["core"]["analyst_email"],
    )
    g.add_analyst(analyst)

    # create Userdata dict
    g.userdata = {
        "notes": obj["metadata"]["core"]["notes"],
        "description": obj["metadata"]["core"]["description"],
        "title": obj["metadata"]["core"]["title"],
        "sadfaceVersion": obj["metadata"]["core"]["version"],
    }

    return g


def atom_from_sadface(obj: sadface.Node, config: ConverterConfig) -> AtomNode:
    """Generate AtomNode object from SADFace Node object."""
    timestamp = pendulum.now()
    return config.AtomNodeClass(
        id=obj["id"],
        text=utils.parse(obj["text"], config.nlp),
        userdata=obj["metadata"],
        metadata=config.MetadataClass(timestamp, timestamp),
    )


def scheme_from_sadface(obj: sadface.Node, config: ConverterConfig) -> SchemeNode:
    """Generate SchemeNode object from SADFace Node object."""
    name = None

    if obj["name"] == "support":
        name = Support.DEFAULT
    elif obj["name"] == "attack":
        name = Attack.DEFAULT
    elif obj["name"] == "rephrase":
        name = Rephrase.DEFAULT
    elif obj["name"] == "preference":
        name = Preference.DEFAULT

    timestamp = pendulum.now()

    return config.SchemeNodeClass(
        id=obj["id"],
        userdata=obj["metadata"],
        metadata=config.MetadataClass(timestamp, timestamp),
        scheme=name,
    )


def edge_from_sadface(
    obj: sadface.Edge, nodes: t.Mapping[str, AbstractNode], config: ConverterConfig
) -> t.Optional[Edge]:
    """Generate Edge object from SADFace Edge format."""
    source_id = obj.get("source_id")
    target_id = obj.get("target_id")

    if source_id in nodes and target_id in nodes:
        return config.EdgeClass(
            id=obj["id"],
            source=nodes[source_id],
            target=nodes[target_id],
        )
    else:
        warn_missing_nodes(obj["id"], source_id, target_id)

    return None



import re
import typing as t

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.models.graph import Graph
from arguebuf.models.node import (
    AtomNode,
    Attack,
    Rephrase,
    Support,
)
from arguebuf.services import utils


def from_kialo(
    obj: t.TextIO,
    name: t.Optional[str] = None,
    config: ConverterConfig = DefaultConverter,
) -> Graph:
    if name_match := re.search(r"Discussion Title: (.*)", obj.readline()):
        name = name_match[1]

    # After the title, an empty line should follow
    assert obj.readline().strip() == ""

    g = config.GraphClass(name)

    # Example: 1.1. Pro: Gold is better than silver.
    # Pattern: {ID}.{ID}. {STANCE (OPTIONAL)}: {TEXT}
    pattern = re.compile(r"^(1\.(?:\d+\.)+) (?:(Con|Pro): )?(.*)")
    current_line = obj.readline()
    next_line = obj.readline()

    mc_match = re.search(r"^((?:\d+\.)+) (.*)", current_line)

    if not mc_match:
        raise ValueError("The major claim is not present in the third line!")

    mc_id = mc_match[1]
    mc_text = mc_match[2]

    # See in the following while loop for explanation of this block
    while next_line and not pattern.search(next_line):
        mc_text = f"{mc_text}\n{next_line.strip()}"
        next_line = obj.readline()

    mc = _kialo_atom_node(mc_id, mc_text, config.nlp, config.AtomNodeClass)
    g.add_node(mc)
    g.major_claim = mc

    current_line = next_line
    next_line = obj.readline()

    while current_line:
        if current_match := pattern.search(current_line):
            source_id = current_match[1]
            source_id_parts = source_id[:-1].split(".")
            level = len(source_id_parts)
            stance = current_match[2]
            text = current_match[3]

            # The text of a node is allowed to span multiple lines.
            # Thus, we need to look ahead to concatenate the complete text.
            # As long as the pattern is not found in the next line,
            # we assume that the text belongs to the previous statement.
            while next_line and not pattern.search(next_line):
                text = f"{text}\n{next_line.strip()}"
                next_line = obj.readline()

            assert source_id
            assert text

            if id_ref_match := re.search(r"^-> See ((?:\d+\.)+)", text):
                id_ref = id_ref_match[1]
                source = g.atom_nodes[id_ref]
            else:
                source = _kialo_atom_node(
                    source_id, text, config.nlp, config.AtomNodeClass
                )
                g.add_node(source)

            if stance:
                stance = stance.lower()
                scheme = config.SchemeNodeClass(
                    Attack.DEFAULT if stance == "con" else Support.DEFAULT,
                    id=f"{source_id}scheme",
                )
            else:
                scheme = config.SchemeNodeClass(
                    Rephrase.DEFAULT, id=f"{source_id}scheme"
                )

            target_id = ".".join(source_id_parts[:-1] + [""])
            target = g.atom_nodes[target_id]

            g.add_node(scheme)
            g.add_edge(config.EdgeClass(source, scheme, id=f"{source.id}->{scheme.id}"))
            g.add_edge(config.EdgeClass(scheme, target, id=f"{scheme.id}->{target.id}"))

            current_line = next_line
            next_line = obj.readline()

    return g


def _kialo_atom_node(
    id: str,
    text: str,
    nlp: t.Optional[t.Callable[[str], t.Any]],
    atom_class: t.Type[AtomNode],
) -> AtomNode:
    # Remove backslashes before parentheses/brackets
    text = re.sub(r"\\([\[\]\(\)])", r"\1", text)

    # Remove markdown links
    text = re.sub(
        r"\[(.*?)\]\(.*?\)",
        r"\1",
        text,
    )

    # Apply user-provided nlp function
    text = utils.parse(text, nlp)

    return atom_class(text, id=id)



import typing as t

from arguebuf.models.node import AbstractNode, AtomNode, SchemeNode

_Node = t.TypeVar("_Node", AtomNode, SchemeNode, AbstractNode)

# https://eddmann.com/posts/depth-first-search-and-breadth-first-search-in-python/
def dfs(
    start: _Node,
    connections: t.Callable[[AbstractNode], t.AbstractSet[_Node]],
    include_start: bool = True,
) -> t.List[_Node]:
    # Need to use a dict since a set does not preserve order
    visited: dict[_Node, None] = {}
    stack: list[_Node] = [start]

    while stack:
        node = stack.pop()

        if node not in visited:
            visited[node] = None
            stack.extend(set(connections(node)) - visited.keys())

    if include_start:
        return list(visited.keys())

    return [node for node in visited if node != start]


def bfs(
    start: _Node,
    connections: t.Callable[[AbstractNode], t.AbstractSet[_Node]],
    include_start: bool = True,
) -> t.List[_Node]:
    # Need to use a dict since a set does not preserve order
    visited: dict[_Node, None] = {}
    queue: list[_Node] = [start]

    while queue:
        node = queue.pop(0)

        if node not in visited:
            visited[node] = None
            queue.extend(set(connections(node)) - visited.keys())

    if include_start:
        return list(visited.keys())

    return [node for node in visited if node != start]


def node_distance(
    start_node: AbstractNode,
    end_node: AbstractNode,
    connections: t.Callable[[AbstractNode], t.Iterable[AbstractNode]],
    max_distance: t.Optional[int] = None,
    directed: bool = True,
) -> t.Optional[int]:
    """Get the distance between `start_node` and `end_node` in the graph.

    Args:
        start_node: Node object that is part of the graph.
        end_node: Node object that is part of the graph.
        connections: Callable containing links between nodes
            (e.g., `graph.incoming_atom_nodes` and `graph.outgoing_nodes`)
        max_distance: Only search for nodes having at most a distance of this argument.
            Especially helpful when dealing with large graphs where shorts paths are searched for.
        directed: If `False`, also search for the direction `end_node` -> `start_node`.

    Returns:
        Number of nodes in between or `None` if no path between

    Examples:
        >>> from arguebuf import Graph, AtomNode, SchemeNode, Edge
        >>> g = Graph()
        >>> n1 = AtomNode("Premise")
        >>> n2 = SchemeNode()
        >>> n3 = AtomNode("Claim")
        >>> e1 = Edge(n1, n2)
        >>> e2 = Edge(n2, n3)
        >>> g.add_node(n1)
        >>> g.add_node(n2)
        >>> len(g.nodes)
        2
        >>> g.add_edge(e1)
        >>> g.add_edge(e2)
        >>> len(g.edges)
        2
        >>> node_distance(n1, n3, g.outgoing_nodes)
        2
        >>> node_distance(n3, n1, g.outgoing_nodes)
    """

    if start_node == end_node:
        return 0

    dist = _directed_node_distance(start_node, end_node, connections, max_distance)

    if dist is None and not directed:
        dist = _directed_node_distance(end_node, start_node, connections, max_distance)

    return dist


def _directed_node_distance(
    start_node: AbstractNode,
    end_node: AbstractNode,
    connections: t.Callable[[AbstractNode], t.Iterable[AbstractNode]],
    max_distance: t.Optional[int] = None,
) -> t.Optional[int]:
    expansion: t.List[t.Tuple[AbstractNode, int]] = [
        (n, 1) for n in connections(start_node)
    ]

    while len(expansion) > 0:
        candidate, distance = expansion.pop()

        if max_distance is not None and distance > max_distance:
            continue
        elif candidate == end_node:
            return distance
        else:
            expansion.extend((n, distance + 1) for n in connections(candidate))

    return None



from __future__ import annotations

import typing as t

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.converters.from_aml import from_aml
from arguebuf.converters.from_brat import from_brat
from arguebuf.converters.from_json import from_json
from arguebuf.converters.from_kialo import from_kialo
from arguebuf.converters.from_microtexts import from_microtexts
from arguebuf.models.graph import Graph


def from_io(
    obj: t.TextIO,
    suffix: str,
    name: t.Optional[str] = None,
    config: ConverterConfig = DefaultConverter,
) -> Graph:
    """Generate Graph structure from IO argument graph file(Link?)."""

    if suffix == ".ann":
        return from_brat(obj, name, config)
    if suffix == ".txt":
        return from_kialo(obj, name, config)
    if suffix == ".aml":
        return from_aml(obj, name, config)
    if suffix == ".xml":
        return from_microtexts(obj, name, config)

    return from_json(obj, name, config)



from __future__ import annotations

import typing as t

from arguebuf.converters.config import GraphFormat
from arguebuf.converters.to_json import to_json
from arguebuf.models.graph import Graph


def to_io(
    graph: Graph,
    obj: t.TextIO,
    format: GraphFormat = GraphFormat.ARGUEBUF,
    pretty: bool = False,
) -> None:
    """Export structure of Graph instance to IO argument graph format."""
    to_json(graph, obj, format, pretty)



from __future__ import annotations

import typing as t

from arg_services.graph.v1 import graph_pb2
from google.protobuf.json_format import ParseDict

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.converters.from_aif import from_aif
from arguebuf.converters.from_ova import from_ova
from arguebuf.converters.from_protobuf import from_protobuf
from arguebuf.models.graph import Graph
from arguebuf.schemas import aif, ova


def from_dict(
    obj: t.Mapping[str, t.Any],
    name: t.Optional[str] = None,
    config: ConverterConfig = DefaultConverter,
) -> Graph:
    """Generate Graph structure from DICT argument graph file(Link?)."""
    if "analysis" in obj:
        return from_ova(t.cast(ova.Graph, obj), name, config)

    if "locutions" in obj:
        return from_aif(t.cast(aif.Graph, obj), name, config)

    return from_protobuf(ParseDict(obj, graph_pb2.Graph()), name, config)



from __future__ import annotations

import typing as t

from google.protobuf.json_format import MessageToDict

from arguebuf.converters.config import GraphFormat
from arguebuf.converters.to_aif import to_aif
from arguebuf.converters.to_protobuf import to_protobuf
from arguebuf.models.graph import Graph


def to_dict(graph: Graph, format: GraphFormat) -> t.Dict[str, t.Any]:
    """Export structure of Graph instance to DICT argument graph format."""

    if format == GraphFormat.AIF:
        return t.cast(dict[str, t.Any], to_aif(graph))

    return MessageToDict(to_protobuf(graph), including_default_value_fields=False)



from __future__ import absolute_import, annotations

import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass

import pendulum

from arguebuf.models.metadata import Metadata
from arguebuf.models.participant import Participant
from arguebuf.models.reference import Reference
from arguebuf.models.scheme import Attack, Preference, Rephrase, Scheme, Support
from arguebuf.models.typing import TextType
from arguebuf.models.userdata import Userdata
from arguebuf.schemas import argdown
from arguebuf.services import utils

NO_SCHEME_LABEL = "Unknown"


@dataclass
class Color:
    bg: str
    fg: str
    border: str

    def __init__(
        self,
        bg: t.Optional[str] = None,
        fg: t.Optional[str] = None,
        border: t.Optional[str] = None,
    ) -> None:
        self.bg = bg or "#000000"
        self.fg = fg or "#ffffff"
        self.border = border or self.bg


scheme2color: t.Dict[t.Type[Scheme], Color] = {
    Support: Color(bg="#4CAF50"),
    Attack: Color(bg="#F44336"),
    Rephrase: Color(bg="#009688"),
    Preference: Color(bg="#009688"),
}


class AbstractNode(ABC):
    """Node in the AIF format."""

    _id: str
    metadata: Metadata
    userdata: Userdata

    def __init__(
        self,
        metadata: t.Optional[Metadata] = None,
        userdata: t.Optional[Userdata] = None,
        id: t.Optional[str] = None,
    ):
        self._id = id or utils.uuid()
        self.metadata = metadata or Metadata()
        self.userdata = userdata or {}

        self.__post_init__()

    def __post_init__(self):
        pass

    def __repr__(self):
        return utils.class_repr(self, [self._id])

    @property
    def id(self) -> str:
        return self._id

    @property
    @abstractmethod
    def label(self) -> str:
        """Generate a matching node label (e.g., for graphviz)"""

    @abstractmethod
    def color(self, major_claim: bool) -> Color:
        """Get the color used in OVA based on `category`."""


class AtomNode(AbstractNode, t.Generic[TextType]):
    __slots__ = (
        "_id",
        "metadata",
        "userdata",
        "text",
        "_reference",
        "_participant",
    )

    text: TextType
    _reference: t.Optional[Reference]
    _participant: t.Optional[Participant]

    def __init__(
        self,
        text: TextType,
        reference: t.Optional[Reference] = None,
        participant: t.Optional[Participant] = None,
        metadata: t.Optional[Metadata] = None,
        userdata: t.Optional[Userdata] = None,
        id: t.Optional[str] = None,
    ):
        super().__init__(metadata, userdata, id)
        self.text = text
        self._reference = reference
        self._participant = participant

    @property
    def plain_text(self) -> str:
        """Get the standard `text` as string."""

        return utils.xstr(self.text)

    @property
    def label(self) -> str:
        return self.plain_text

    @property
    def reference(self) -> t.Optional[Reference]:
        return self._reference

    @property
    def participant(self) -> t.Optional[Participant]:
        return self._participant

    def __repr__(self):
        return utils.class_repr(self, [self._id, self.plain_text])

    @classmethod
    def from_argdown_json(
        cls,
        obj: argdown.Node,
        nlp: t.Optional[t.Callable[[str], t.Any]] = None,
    ) -> AtomNode:
        """Generate AtomNode object from Argdown JSON Node object."""
        timestamp = pendulum.now()
        return cls(
            id=obj["id"],
            text=utils.parse(obj["labelText"], nlp),
            metadata=Metadata(timestamp, timestamp),
        )

    def color(self, major_claim: bool) -> Color:
        """Get the color for rendering the node."""
        if major_claim:
            return Color(bg="#0D47A1")

        return Color(bg="#2196F3")


class SchemeNode(AbstractNode):
    __slots__ = (
        "_id",
        "created",
        "updated",
        "userdata",
        "scheme",
        "premise_descriptors",
    )

    scheme: t.Optional[Scheme]
    premise_descriptors: t.List[str]

    def __init__(
        self,
        scheme: t.Optional[Scheme] = None,
        premise_descriptors: t.Optional[t.List[str]] = None,
        metadata: t.Optional[Metadata] = None,
        userdata: t.Optional[Userdata] = None,
        id: t.Optional[str] = None,
    ):
        super().__init__(metadata, userdata, id)
        self.scheme = scheme
        self.premise_descriptors = premise_descriptors or []

    def __repr__(self):
        return utils.class_repr(
            self,
            [
                self._id,
                type(self.scheme).__name__ if self.scheme else NO_SCHEME_LABEL,
                self.scheme.value if self.scheme else "",
            ],
        )

    @property
    def label(self) -> str:
        label = NO_SCHEME_LABEL

        if self.scheme:
            label = type(self.scheme).__name__

            if self.scheme.value != "Default":
                label = f"{label}: {self.scheme.value}"

        return label

    def color(self, major_claim: bool) -> Color:
        """Get the color used in OVA based on `category`."""

        return scheme2color[type(self.scheme)] if self.scheme else Color(bg="#009688")


AtomOrSchemeNode = t.Union[AtomNode, SchemeNode]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['arguebuf',
 'arguebuf.cli',
 'arguebuf.converters',
 'arguebuf.models',
 'arguebuf.schemas',
 'arguebuf.services']

package_data = \
{'': ['*']}

install_requires = \
['arg-services>=1.0.8,<2.0.0',
 'graphviz>=0.20.1,<0.21.0',
 'lxml>=4.9.2,<5.0.0',
 'networkx>=2.8.8,<4.0',
 'pendulum>=2.1.2,<3.0.0']

extras_require = \
{'cli': ['typer[all]>=0.7.0,<0.8.0',
         'deepl>=1.12.0,<2.0.0',
         'multimethod>=1.9.1,<2.0.0'],
 'docs': ['pdoc>=12.3.1,<13.0.0'],
 'graphviz': ['pygraphviz>=1.10,<2.0']}

entry_points = \
{'console_scripts': ['arguebuf = arguebuf.cli.app:cli']}

setup_kwargs = {
    'name': 'arguebuf',
    'version': '2.0.0b12',
    'description': 'A library for loading argument graphs in various formats (e.g., AIF).',
    'long_description': '# Arguebuf\n\nArguebuf is a format for serializing argument graphs and specified using Protobuf.\nThe complete specification and documentation is available at the [Buf Schema Registry](https://buf.build/recap/arg-services/docs/main:arg_services.graph.v1).\nWhile Protobuf automatically generates native code for all major programming languages (including Python), we created a custom implementation that provides some additional benefits, including:\n\n- The ability to import existing formats like [AIF](http://www.argumentinterchange.org), [SADFace](https://github.com/Open-Argumentation/SADFace), and a few others.\n- Export of Arguebuf graphs to AIF, [NetworkX](https://networkx.org), and [Graphviz](https://graphviz.org).\n- Integration with the popular NLP library [spaCy](http://spacy.io).\n- Various helper methods to programmatically manipulate/create argument graphs.\n- More pythonic interfaces than the regular code generated by `protoc`.\n\nYou can easily install the library from [PyPI](https://pypi.org/project/arguebuf/) using pip. The documentation is hosted on [ReadTheDocs](https://arguebuf.readthedocs.io/en/latest/)\n\n## Command Line Interface (CLI)\n\nWe also offer some tools to simplify dealing with structured argument graphs.\nAmong others, it is possible to convert graphs between different formats, translate them, and render images using graphviz.\nTo use it, install the `cli` extras when installing the package.\nWhen using `pip`, this can be accomplished with\n\n`pip install arguebuf[cli]`\n\nAfterwards, you can execute it by calling `arguebuf`, for example:\n\n`arguebuf --help`\n\n## Theoretical Foundations\n\nAn argument graph is way to represent _structured_ argumentation.\nWhat sets it apart from _unstructured_ argumentation (e.g., in newspaper articles) is that only the essential/argumentative parts of a text art part of this representation.\nThese _units_ of argumentation are also called ADUs (argumentative discourse units).\nThe length of ADUs can differ dramatically (depending on various factors like the context), meaning they might contain only a few words, a sentence, or even a whole paragraph.\nThe structure of an argument is then represented through _relations_ between these units.\nFor this purpose, they can be further subdivided into _claims_ and _premises_:\nA claim is a statement that is supported or attacked by one or multiple premises.\nAt the same time, a claim may also function as a premise for another claim, making it possible to construct even complex _argument graphs_.\nOne of the claims is called _major claim_ and represents the overall claim of the whole argument.\nIn many cases (but not all), this major claim is located right at the top of the graph\nHere is a rather simple example.\n\nTODO: Example\n\nClaims, premises and the major claim are represented as _atom nodes_ while relations between them are represented by _scheme nodes_.\nThe set of nodes $V = A \\cup S$ is composed of the set of atom nodes $A$ and the set of scheme nodes $S$.\nThe supporting or attacking relations are encoded in a set of edges $E \\subseteq V \\times V \\setminus A \\times A$.\nEdges can be drawn between any type of nodes except for two atom nodes.\nFor instance, it is possible to connect two scheme nodes to support or attack the _inference_ between two ADUs.\nBased on this, we define an argument graph $G$ as the triple $G = ( V , E , M )$.\n\n## User Guide\n\nWhen importing this library, we recommend using an abbreviation such as `ag` (for _argument graph_).\n\n```python\nimport arguebuf as ag\n```\n\nIn the following, we will introduce the most important features of `arguebuf`.\nFor more examples, we encourage you to visit our [API documentation](./api.md) which contains examples for most of the methods.\n\n### Importing an Existing Graph\n\nWe support multiple established formats to represent argument graphs: `AIF`, `OVA`, `SADFace`, `ArgDown`, `BRAT`, and `Kialo`.\nGiven an input file, the library can automatically determine the correct format and convert it to a representation in the `arguebuf` format.\nOne can either pass a string pointing to the file or a `pathlib.Path` object.\n\n```python\ngraph = ag.from_file("graph.json")\n```\n\nIt is also possible to load multiple graphs within a folder.\nHere, you need to pass the folder along with a [glob pattern](https://docs.python.org/3/library/fnmatch.html#module-fnmatch) for selecting the argument graphs.\nThis also enables to recursively load all argument graphs from a common parent directory.\n\n```python\ngraphs = ag.from_folder("./data", "**/*.json")\n```\n\nSince atom nodes contain textual information that may need to be analyzed using NLP techniques, we support passing a custom `nlp` function to these loader methods.\nThis also makes it really easy to integrate the popular [`spacy` library](http://spacy.io) with `arguebuf` as all texts of atom nodes are automatically converted to a spacy `Doc`.\n\n```python\nimport spacy\nnlp = spacy.load("en_core_web_lg")\ngraph = ag.from_file("graph.json", nlp=nlp)\n```\n\n### Programmatically Create a New Graph\n\nInstead of importing an existing graph, you can also create a new one using an object-oriented API using our library.\nTo illustrate this, we generate a graph with two premises that are connected to a major claim.\n**Please note:** In case edges with nodes not yet contained in the graph are added, the respective nodes are added automatically.\n\n```python\ngraph = ag.Graph("TITLE")\n\npremise1 = ag.AtomNode("Text of premise 1")\npremise2 = ag.AtomNode("Text of premise 2")\nclaim = ag.AtomNode("Text of claim")\n\nscheme1 = ag.SchemeNode(ag.Support.DEFAULT)\nscheme2 = ag.SchemeNode(ag.Attack.DEFAULT)\n\ngraph.add_edge(ag.Edge(premise1, scheme1))\ngraph.add_edge(ag.Edge(scheme1, claim))\ngraph.add_edge(ag.Edge(premise2, scheme2))\ngraph.add_edge(ag.Edge(scheme2, claim))\n\ngraph.major_claim = claim\n```\n\nWith this code, we get the following output\n\n![Output of programmatic graph creation](./img/programmatic.png)\n\n### Exporting Argument Graphs\n\nWe support different output formats and integration with other libraries to ease the use of argument graphs.\nHave a look at the following code snippet to get an overview of the possibilities\n\n```python\n# Export to graphviz DOT format\ndot = ag.to_graphviz(graph)\n\n# Export an image of this dot source to a file\nag.render(dot, "./graph.pdf")\n\n# Convert to NetworkX graph\nnx = ag.to_networkx(graph)\n\n# Save the graph as Arguebuf\nag.to_file(graph, "./graph.json")\n\n# Save the graph as AIF\nag.to_file(graph, "./graph.json", ag.GraphFormat.AIF)\n```\n',
    'author': 'Mirko Lenz',
    'author_email': 'info@mirko-lenz.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://recap.uni-trier.de',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import typing as t
from dataclasses import dataclass

from lxml import etree

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.models.graph import Graph
from arguebuf.models.node import Attack, Support
from arguebuf.schemas.microtexts import EdgeType
from arguebuf.services import utils


@dataclass
class _Edge:
    source: str
    target: str
    type: EdgeType


def transform_edges(elems: t.Iterable[t.Any]) -> dict[str, _Edge]:
    return {
        str(elem.attrib["id"]): _Edge(
            str(elem.attrib["src"]),
            str(elem.attrib["trg"]),
            EdgeType(str(elem.attrib["type"])),
        )
        for elem in elems
        if isinstance(elem, etree._Element)
    }


def from_microtexts(
    obj: t.IO, name: t.Optional[str] = None, config: ConverterConfig = DefaultConverter
) -> Graph:
    """
    Generate Graph structure from AML argument graph file
    ElementTree XML API: https://docs.python.org/3/library/xml.etree.elementtree.html#
    """

    tree = etree.parse(obj)
    root = tree.getroot()
    id = root.get("id")
    g = config.GraphClass(name or id)
    g.userdata = {"stance": root.get("stance"), "topic": root.get("topic_id")}

    segmentation_edge_tags = root.xpath("//edge[@type='seg']")
    atom_edge_tags = root.xpath("//edge[@type='sup' or @type='exa' or @type='reb']")
    scheme_edge_tags = root.xpath("//edge[@type='add']")
    edge_edge_tags = root.xpath("//edge[@type='und']")
    edu_tags = root.xpath("//edu")
    joint_tags = root.xpath("//joint")
    adu_tags = root.xpath("//adu")
    assert (
        isinstance(edu_tags, list)
        and isinstance(joint_tags, list)
        and isinstance(adu_tags, list)
        and isinstance(segmentation_edge_tags, list)
        and isinstance(atom_edge_tags, list)
        and isinstance(edge_edge_tags, list)
        and isinstance(scheme_edge_tags, list)
    )

    source_tags = {
        str(elem.attrib["id"]): elem
        for elem in [*edu_tags, *joint_tags]
        if isinstance(elem, etree._Element)
    }
    adu2source = {
        edge.target: edge.source
        for edge in transform_edges(segmentation_edge_tags).values()
    }
    atom_edges = transform_edges(atom_edge_tags)
    scheme_edges = transform_edges(scheme_edge_tags)
    edge_edges = transform_edges(edge_edge_tags)

    for adu in adu_tags:
        if isinstance(adu, etree._Element) and (adu_id := adu.get("id")):
            adu_source_id = adu2source[adu_id]
            adu_source = source_tags[adu_source_id]

            if adu_source.text is not None:
                atom = config.AtomNodeClass(
                    utils.parse(adu_source.text, config.nlp), id=adu_id
                )

                if type := adu.get("type"):
                    atom.userdata["type"] = type

                g.add_node(atom)

                if adu_source.get("implicit"):
                    g.major_claim = atom

    for edge_id, edge in atom_edges.items():
        scheme_node = config.SchemeNodeClass(id=edge_id)

        if edge.type == EdgeType.SUPPORT_DEFAULT:
            scheme_node.scheme = Support.DEFAULT
        elif edge.type == EdgeType.SUPPORT_EXAMPLE:
            scheme_node.scheme = Support.EXAMPLE
        elif edge.type == EdgeType.ATTACK_ATOM:
            scheme_node.scheme = Attack.DEFAULT

        if (source_atom := g.atom_nodes.get(edge.source)) and (
            target_atom := g.atom_nodes.get(edge.target)
        ):
            g.add_edge(config.EdgeClass(source_atom, scheme_node))
            g.add_edge(config.EdgeClass(scheme_node, target_atom))

    for edge in scheme_edges.values():
        if (scheme_node := g.scheme_nodes.get(edge.target)) and (
            atom_node := g.atom_nodes.get(edge.source)
        ):
            g.add_edge(config.EdgeClass(atom_node, scheme_node))

    for edge_id, edge in edge_edges.items():
        if (target_scheme := g.scheme_nodes.get(edge.target)) and (
            atom_node := g.atom_nodes.get(edge.source)
        ):
            source_scheme = config.SchemeNodeClass(id=edge_id)
            g.add_edge(config.EdgeClass(atom_node, source_scheme))
            g.add_edge(config.EdgeClass(source_scheme, target_scheme))

    return g



import typing as t
import xml.etree.ElementTree as ET

import pendulum

from arguebuf.converters.config import ConverterConfig, DefaultConverter
from arguebuf.models.edge import Edge
from arguebuf.models.graph import Graph
from arguebuf.models.metadata import Metadata
from arguebuf.models.node import AtomNode, Attack, SchemeNode, Support
from arguebuf.services import utils


def from_aml(
    obj: t.IO, name: t.Optional[str] = None, config: ConverterConfig = DefaultConverter
) -> Graph:
    """
    Generate Graph structure from AML argument graph file
    ElementTree XML API: https://docs.python.org/3/library/xml.etree.elementtree.html#
    """

    tree = ET.parse(obj)
    root = tree.getroot()

    # create nodes and edges from AU element
    au = root.find("AU")
    g = config.GraphClass(name)
    assert isinstance(au, ET.Element)
    g = read_au(au, g, config)

    # create edge objects and add to g

    # create Metadata object
    # create Analyst object
    # create Userdata

    return g


def read_au(au: ET.Element, g: Graph, config: ConverterConfig):
    # create the conclusion node
    prop = au.find("PROP")
    if prop is not None:
        conclusion_node = atom_from_aml(prop, config)
        if conclusion_node.id not in g.atom_nodes:
            g.add_node(conclusion_node)

        refutation = au.find("REFUTATION")
        if refutation is not None:
            # handle refutation
            # consider, that refutations are argument units (AU's)
            read_refutation(conclusion_node, g, refutation, config)

        # read premises and store in a list (list of ET.Element objects)
        premises = [elem for elem in au if elem.tag == "CA" or elem.tag == "LA"]
        for elem in premises:
            if elem.tag == "CA":
                read_ca(conclusion_node, g, elem, config)
            elif elem.tag == "LA":
                read_la(conclusion_node, g, elem, config)

    # return graph
    return g


def read_refutation(
    conclusion: AtomNode, g, refutation: ET.Element, config: ConverterConfig
):
    # consists of one AU
    # get refutation
    au = refutation.find("AU")

    if au is not None:
        prop = au.find("PROP")

        if prop is not None:
            premise_node = atom_from_aml(prop, config)
            g.add_node(premise_node)

            # create SchemeNode (Attack)
            scheme_node = scheme_from_aml(prop, config, refutation=True)
            g.add_node(scheme_node)

            # create edge from premise to schemeNode
            g.add_edge(Edge(premise_node, scheme_node))

            # create edge from schemeNode to conclusion
            g.add_edge(Edge(scheme_node, conclusion))

            # read the rest of au
            # read premises and store in a list (list of ET.Element objects)
            premises = [elem for elem in au if elem.tag == "CA" or elem.tag == "LA"]
            for elem in premises:
                if elem.tag == "CA":
                    read_ca(premise_node, g, elem, config)
                elif elem.tag == "LA":
                    read_la(premise_node, g, elem, config)


def read_ca(conclusion: AtomNode, g: Graph, ca: ET.Element, config: ConverterConfig):
    # first read premises
    for au in ca:
        # get premise
        prop = au.find("PROP")

        if prop is not None:
            premise_node = atom_from_aml(prop, config)
            g.add_node(premise_node)

            # create SchemeNode
            scheme_node = scheme_from_aml(prop, config)
            g.add_node(scheme_node)

            # create edge from premise to schemeNode
            g.add_edge(config.EdgeClass(premise_node, scheme_node))

            # create edge from schemeNode to conclusion
            g.add_edge(config.EdgeClass(scheme_node, conclusion))

            # read the rest of au
            read_au(au, g, config)


def read_la(conclusion: AtomNode, g: Graph, la: ET.Element, config: ConverterConfig):
    # first read premises
    for au in la:
        # get premise
        prop = au.find("PROP")

        if prop is not None:
            premise_node = atom_from_aml(prop, config)
            g.add_node(premise_node)

            # create SchemeNode
            scheme_node = scheme_from_aml(prop, config)
            g.add_node(scheme_node)

            # create edge from premise to schemeNode
            g.add_edge(Edge(premise_node, scheme_node))

            # create edge from schemeNode to conclusion
            g.add_edge(Edge(scheme_node, conclusion))

            # read the rest of au
            read_au(au, g, config)


def atom_from_aml(
    obj: ET.Element,
    config: ConverterConfig,
) -> AtomNode:
    """
    Generate Node object from AML Node format. obj is a AML "PROP" element.
    """
    # get id of PROP
    id = obj.get("identifier")

    # read text of PROP
    text = None
    text_element = obj.find("PROPTEXT")

    if text_element is not None:
        text = text_element.text

    # read owners of PROP
    owner_list = obj.findall("OWNER")
    owners_lst = []
    if not owner_list:
        # if not empty, do something
        for owner in owner_list:
            owners_lst.append(owner.get("name"))
        owners = {"owners": ", ".join(owners_lst)}
    else:
        owners = {}

    # create timestamp
    timestamp = pendulum.now()

    return config.AtomNodeClass(
        id=id,
        text=utils.parse(text, config.nlp),
        metadata=Metadata(timestamp, timestamp),
        userdata=owners,
    )


def scheme_from_aml(
    obj: ET.Element,
    config: ConverterConfig,
    refutation=False,
) -> SchemeNode:
    """Generate SchemeNode object from AML Node format. obj is a AML "PROP" element."""

    # get id of PROP
    if "identifier" in obj.attrib:
        id = obj.get("identifier")
    else:
        id = None

    # read owners of PROP
    owner_list = obj.findall("OWNER")
    owners_lst = []
    if not owner_list:
        # if not empty, do something
        for owner in owner_list:
            owners_lst.append(owner.get("name"))
        owners = {"owners": ", ".join(owners_lst)}
    else:
        owners = {}

    # create timestamp
    timestamp = pendulum.now()

    # get scheme name
    scheme = None
    if refutation:
        scheme = Attack.DEFAULT
    else:
        inscheme = obj.find("INSCHEME")
        if inscheme is not None:  # if INSCHEME element is available
            # get scheme
            aml_scheme = inscheme.attrib["scheme"]
            contains_scheme = False
            for supp_scheme in Support:
                if supp_scheme.value.lower().replace(
                    " ", ""
                ) in aml_scheme.lower().replace(" ", ""):
                    scheme = supp_scheme
                    contains_scheme = True
                    break
            if not contains_scheme:
                scheme = Support.DEFAULT
        else:  # if INSCHEME element is not available
            scheme = Support.DEFAULT

    return config.SchemeNodeClass(
        metadata=Metadata(timestamp, timestamp),
        scheme=scheme,
        userdata=owners,
        id=id,
    )



"""
.. include:: ../README.md
"""

import logging

from . import schemas
from .converters import *
from .models import *
from .services import *

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = (
    # models
    "Graph",
    "AbstractNode",
    "AtomNode",
    "SchemeNode",
    "Edge",
    "Analyst",
    "Metadata",
    "Participant",
    "Reference",
    "Resource",
    "Attack",
    "Preference",
    "Rephrase",
    "Scheme",
    "Support",
    "Userdata",
    "AtomOrSchemeNode",
    # converters
    "ConverterConfig",
    "GraphFormat",
    "copy",
    "from_aif",
    "from_aml",
    "from_argdown",
    "from_brat",
    "CasebaseFilter",
    "from_casebase",
    "from_dict",
    "from_io",
    "from_json",
    "from_kialo",
    "from_microtexts",
    "from_ova",
    "from_file",
    "from_folder",
    "from_protobuf",
    "from_sadface",
    "to_aif",
    "to_dict",
    "render",
    "to_graphviz",
    "to_io",
    "to_json",
    "to_networkx",
    "to_file",
    "to_folder",
    "to_protobuf",
    # services
    "bfs",
    "dfs",
    "node_distance",
    "uuid",
    # schemas
    "schemas",
)



from __future__ import annotations

import typing as t

import pendulum
from pendulum.datetime import DateTime



class Metadata:
    created: DateTime
    updated: DateTime
    # _analyst: t.Optional[Analyst] = None

    def __init__(
        self,
        created: t.Optional[DateTime] = None,
        updated: t.Optional[DateTime] = None,
    ) -> None:
        now = pendulum.now()

        self.created = created or now
        self.updated = updated or now

    # @property
    # def analyst(self) -> t.Optional[Analyst]:
    #     return self._analyst

    def update(self) -> None:
        self.updated = pendulum.now()



from typing import TypeVar

TextType = TypeVar("TextType")
# AbstractNodeType = TypeVar("AbstractNodeType", bound=AbstractNode)
# AnalystType = TypeVar("AnalystType", bound=Analyst)
# AtomNodeType = TypeVar("AtomNodeType", bound=AtomNode)
# EdgeType = TypeVar("EdgeType", bound=Edge)
# MetadataType = TypeVar("MetadataType", bound=Metadata)
# ParticipantType = TypeVar("ParticipantType", bound=Participant)
# ReferenceType = TypeVar("ReferenceType", bound=Reference)
# ResourceType = TypeVar("ResourceType", bound=Resource)
# SchemeNodeType = TypeVar("SchemeNodeType", bound=SchemeNode)
# UserdataType = TypeVar("UserdataType", bound=Mapping)



from __future__ import absolute_import, annotations

import itertools
import logging
import typing as t

from arguebuf.models.analyst import Analyst
from arguebuf.models.edge import Edge
from arguebuf.models.metadata import Metadata
from arguebuf.models.node import AbstractNode, AtomNode, SchemeNode
from arguebuf.models.participant import Participant
from arguebuf.models.resource import Resource
from arguebuf.models.typing import TextType
from arguebuf.models.userdata import Userdata
from arguebuf.services import traversal, utils
from arguebuf.services.utils import ImmutableDict, ImmutableSet

log = logging.getLogger(__name__)


# noinspection PyProtectedMember
class Graph(t.Generic[TextType]):
    """Graph in AIF format.

    No attribute is mandatory.
    All nodes and edges attributes are read-only.
    """

    __slots__ = (
        "name",
        "_nodes",
        "_atom_nodes",
        "_scheme_nodes",
        "_edges",
        "_incoming_nodes",
        "_incoming_edges",
        "_outgoing_nodes",
        "_outgoing_edges",
        "_major_claim",
        "_resources",
        "_participants",
        "_analysts",
        "metadata",
        "userdata",
        "library_version",
        "schema_version",
    )

    name: str
    _nodes: ImmutableDict[str, AbstractNode]
    _atom_nodes: ImmutableDict[str, AtomNode]
    _scheme_nodes: ImmutableDict[str, SchemeNode]
    _edges: ImmutableDict[str, Edge]
    _incoming_nodes: ImmutableDict[AbstractNode, ImmutableSet[AbstractNode]]
    _incoming_edges: ImmutableDict[AbstractNode, ImmutableSet[Edge]]
    _outgoing_nodes: ImmutableDict[AbstractNode, ImmutableSet[AbstractNode]]
    _outgoing_edges: ImmutableDict[AbstractNode, ImmutableSet[Edge]]
    _resources: ImmutableDict[str, Resource]
    _participants: ImmutableDict[str, Participant]
    _major_claim: t.Optional[AtomNode]
    _analysts: ImmutableDict[str, Analyst]
    library_version: t.Optional[str]
    schema_version: t.Optional[int]
    metadata: Metadata
    userdata: Userdata

    @property
    def edges(self) -> t.Mapping[str, Edge]:
        return self._edges

    @property
    def nodes(self) -> t.Mapping[str, AbstractNode]:
        return self._nodes

    @property
    def atom_nodes(self) -> t.Mapping[str, AtomNode]:
        return self._atom_nodes

    @property
    def scheme_nodes(self) -> t.Mapping[str, SchemeNode]:
        return self._scheme_nodes

    def incoming_nodes(
        self, node: t.Union[str, AbstractNode]
    ) -> t.AbstractSet[AbstractNode]:
        if isinstance(node, str):
            node = self._nodes[node]

        return self._incoming_nodes[node]

    def incoming_atom_nodes(
        self, node: t.Union[str, AbstractNode]
    ) -> t.AbstractSet[AtomNode]:
        if isinstance(node, str):
            node = self._nodes[node]

        incoming_nodes = list(self._incoming_nodes[node])
        incoming_atom_nodes: set[AtomNode] = set()

        while incoming_nodes:
            incoming_node = incoming_nodes.pop()

            if isinstance(incoming_node, AtomNode):
                incoming_atom_nodes.add(incoming_node)
            else:
                incoming_nodes.extend(self._incoming_nodes[incoming_node])

        return incoming_atom_nodes

    def outgoing_nodes(
        self, node: t.Union[str, AbstractNode]
    ) -> t.AbstractSet[AbstractNode]:
        if isinstance(node, str):
            node = self._nodes[node]

        return self._outgoing_nodes[node]

    def outgoing_atom_nodes(
        self, node: t.Union[str, AbstractNode]
    ) -> t.AbstractSet[AtomNode]:
        if isinstance(node, str):
            node = self._nodes[node]

        outgoing_nodes = list(self._outgoing_nodes[node])
        outgoing_atom_nodes: set[AtomNode] = set()

        while outgoing_nodes:
            outgoing_node = outgoing_nodes.pop()

            # If it is an Atom, just add it to our result set
            if isinstance(outgoing_node, AtomNode):
                outgoing_atom_nodes.add(outgoing_node)
            # Otherwise, add the outgoing nodes of the current node to the search path
            else:
                outgoing_nodes.extend(self._outgoing_nodes[outgoing_node])

        return outgoing_atom_nodes

    def incoming_edges(self, node: t.Union[str, AbstractNode]) -> t.AbstractSet[Edge]:
        if isinstance(node, str):
            node = self._nodes[node]

        return self._incoming_edges[node]

    def outgoing_edges(self, node: t.Union[str, AbstractNode]) -> t.AbstractSet[Edge]:
        if isinstance(node, str):
            node = self._nodes[node]

        return self._outgoing_edges[node]

    def scheme_between(
        self, premise: AtomNode, claim: AtomNode
    ) -> t.Optional[SchemeNode]:
        candidates = set(self._outgoing_nodes[premise]).intersection(
            self._incoming_nodes[claim]
        )

        if len(candidates) == 1:
            scheme = next(iter(candidates))

            if isinstance(scheme, SchemeNode):
                return scheme

        return None

    @property
    def resources(self) -> t.Mapping[str, Resource]:
        return self._resources

    @property
    def major_claim(self) -> t.Optional[AtomNode]:
        if self._major_claim:
            return self._major_claim

    @major_claim.setter
    def major_claim(self, value: t.Union[str, AtomNode, None]) -> None:
        if isinstance(value, str):
            value = self._atom_nodes[value]
        elif not (value is None or isinstance(value, AtomNode)):
            raise TypeError(utils.type_error(type(value), AtomNode))

        self._major_claim = value
        # self._metadata.update()

    @property
    def root_node(self) -> t.Optional[AtomNode]:
        """If there is a single node with no outgoing edges, return it, otherwise None"""
        candidates = self.root_nodes

        if len(candidates) == 1:
            return next(iter(candidates))

        return None

    @property
    def root_nodes(self) -> t.Set[AtomNode]:
        """Find all nodes with no outgoing edges"""
        return {
            node
            for node in self.atom_nodes.values()
            if len(self.outgoing_nodes(node)) == 0
        }

    @property
    def leaf_nodes(self) -> t.Set[AtomNode]:
        """Find all nodes with no incoming edges"""
        return {
            node
            for node in self.atom_nodes.values()
            if len(self.incoming_nodes(node)) == 0
        }

    @property
    def participants(self) -> t.Mapping[str, Participant]:
        return self._participants

    @property
    def analysts(self) -> t.Mapping[str, Analyst]:
        return self._analysts

    def __init__(self, name: t.Optional[str] = None):
        """Create a graph from scratch."""

        self.name = name or utils.uuid()
        self._nodes = ImmutableDict()
        self._atom_nodes = ImmutableDict()
        self._scheme_nodes = ImmutableDict()
        self._edges = ImmutableDict()
        self._analysts = ImmutableDict()
        self.userdata = {}
        self._resources = ImmutableDict()
        self._participants = ImmutableDict()
        self.metadata = Metadata()
        self._major_claim = None

        self._incoming_nodes = ImmutableDict()
        self._incoming_edges = ImmutableDict()
        self._outgoing_nodes = ImmutableDict()
        self._outgoing_edges = ImmutableDict()
        self.library_version = None
        self.schema_version = None

        self.__post_init__()

    def __post_init__(self):
        pass

    def __repr__(self):
        return utils.class_repr(self, [self.name])

    def add_node(self, node: AbstractNode) -> None:
        """Add a node to the graph.

        Args:
            node: Node object that is not already part of the graph.

        Examples:
            >>> g = Graph()
            >>> g.add_node(AtomNode("Exemplary node"))
            >>> len(g.nodes)
            1
            >>> g.add_node(SchemeNode())
            >>> len(g.nodes)
            2
            >>> g.add_node("Test")
            Traceback (most recent call last):
            TypeError: Expected type '<class 'arguebuf.node.Node'>', but got '<class 'str'>'. Make sure that you are passing the correct method arguments.
        """

        if not isinstance(node, AbstractNode):
            raise TypeError(utils.type_error(type(node), AbstractNode))

        if node.id in self._nodes:
            raise ValueError(utils.duplicate_key_error(self.name, node.id))

        self._nodes._store[node.id] = node

        if isinstance(node, AtomNode):
            self._atom_nodes._store[node.id] = node

            if node.participant and node.participant.id not in self._participants:
                self.add_participant(node.participant)

            if (
                node.reference
                and node.reference.resource
                and node.reference.resource.id not in self._resources
            ):
                self.add_resource(node.reference.resource)

        elif isinstance(node, SchemeNode):
            self._scheme_nodes._store[node.id] = node

        self._incoming_nodes._store[node] = ImmutableSet()
        self._incoming_edges._store[node] = ImmutableSet()
        self._outgoing_nodes._store[node] = ImmutableSet()
        self._outgoing_edges._store[node] = ImmutableSet()

    def remove_node(self, node: AbstractNode) -> None:
        """Remove a node and its corresponding edges from the graph.

        Args:
            node: Node object that is part of the graph.

        Examples:
            >>> g = Graph()
            >>> n1 = AtomNode("Node1")
            >>> n2 = SchemeNode()
            >>> e = Edge(n1, n2)
            >>> g.add_edge(e)
            >>> len(g.nodes)
            2
            >>> len(g.edges)
            1
            >>> g.remove_node(n1)
            >>> len(g.nodes)
            1
            >>> len(g.edges)
            0
            >>> g.remove_node(n1)
            Traceback (most recent call last):
            KeyError: Node not in graph.
        """

        if node.id not in self.nodes:
            raise KeyError(utils.missing_key_error(self.name, node.id))

        del self._nodes._store[node.id]

        if isinstance(node, AtomNode):
            del self._atom_nodes._store[node.id]
        elif isinstance(node, SchemeNode):
            del self._scheme_nodes._store[node.id]

        neighbor_edges = list(self._incoming_edges[node]) + list(
            self._outgoing_edges[node]
        )

        for edge in neighbor_edges:
            self.remove_edge(edge)

        del self._incoming_nodes._store[node]
        del self._incoming_edges._store[node]
        del self._outgoing_nodes._store[node]
        del self._outgoing_edges._store[node]

    def add_edge(self, edge: Edge) -> None:
        """Add an edge and its nodes (if not already added).

        Args:
            edge: Edge object that is not part of the graph.

        Examples:
            >>> g = Graph()
            >>> n1 = AtomNode("Premise")
            >>> n2 = SchemeNode()
            >>> n3 = AtomNode("Claim")
            >>> e1 = Edge(n1, n2)
            >>> e2 = Edge(n2, n3)
            >>> g.add_edge(e1)
            >>> len(g.edges)
            1
            >>> g.add_edge(e2)
            >>> len(g.edges)
            2
        """
        if not isinstance(edge, Edge):
            raise TypeError(utils.type_error(type(edge), Edge))

        if edge.id in self._edges:
            raise ValueError(utils.duplicate_key_error(self.name, edge.id))

        self._edges._store[edge.id] = edge

        if edge.source.id not in self.nodes:
            self.add_node(edge.source)

        if edge.target.id not in self.nodes:
            self.add_node(edge.target)

        self._outgoing_edges[edge.source]._store.add(edge)
        self._incoming_edges[edge.target]._store.add(edge)
        self._outgoing_nodes[edge.source]._store.add(edge.target)
        self._incoming_nodes[edge.target]._store.add(edge.source)

    def remove_edge(self, edge: Edge) -> None:
        """Remove an edge.

        Args:
            edge: Edge object that is part of the graph.

        Examples:
            >>> g = Graph()
            >>> n1 = AtomNode("Node1")
            >>> n2 = SchemeNode()
            >>> e = Edge(n1, n2)
            >>> g.add_edge(e)
            >>> len(g.edges)
            1
            >>> len(g.nodes)
            2
            >>> g.remove_edge(e)
            >>> len(g.edges)
            0
            >>> len(g.nodes)
            2
        """
        if not isinstance(edge, Edge):
            raise TypeError(utils.type_error(type(edge), Edge))

        if edge.id not in self._edges:
            raise KeyError(utils.missing_key_error(self.name, edge.id))

        del self._edges._store[edge.id]

        self._outgoing_edges[edge.source]._store.remove(edge)
        self._incoming_edges[edge.target]._store.remove(edge)
        self._outgoing_nodes[edge.source]._store.remove(edge.target)
        self._incoming_nodes[edge.target]._store.remove(edge.source)

    ##What does a Resource look like?
    def add_resource(self, resource: Resource) -> None:
        """Add a resource.

        Args:
            resource: Resource object that is not part of the graph.

        Examples:
            >>> g = Graph()
            >>> r1 = Resource("Resource1")
            >>> g.add_resource(r1)
            >>> len(g.resources)
            1
        """
        if not isinstance(resource, Resource):
            raise TypeError(utils.type_error(type(resource), Resource))

        if resource.id in self._resources:
            raise ValueError(utils.duplicate_key_error(self.name, resource.id))

        self._resources._store[resource.id] = resource

    def remove_resource(self, resource: Resource) -> None:
        """Add a resource.

        Args:
            resource: Resource object that is part of the graph.

        Examples:
            >>> g = Graph()
            >>> r1 = Resource("Resource1")
            >>> g.add_resource(r1)
            >>> len(g.resources)
            1
            >>> g.remove_resource(r1)
            >>> len(g.resources)
            0
        """
        if not isinstance(resource, Resource):
            raise TypeError(utils.type_error(type(resource), Resource))

        if resource.id not in self._resources:
            raise ValueError(utils.missing_key_error(self.name, resource.id))

        del self._resources._store[resource.id]

        for node in self._atom_nodes.values():
            if node.reference and node.reference.resource == resource:
                node.reference._resource = None
                node.reference.offset = None

    def clean_resources(self) -> None:
        """Remove resources from the graph that are used by no nodes"""

        node_resources = {
            node.reference.resource.id
            for node in self._atom_nodes.values()
            if node.reference and node.reference.resource
        }

        for resource_id in set(self._resources):
            if resource_id not in node_resources:
                del self._resources._store[resource_id]

    def add_participant(self, participant: Participant) -> None:
        """Add a resource.

        Args:
            participant: Participant object that is not part of the graph.

        Examples:
            >>> g = Graph()
            >>> p1 = Participant("Participant1")
            >>> g.add_participant(p1)
            >>> len(g.participants)
            1
        """
        if not isinstance(participant, Participant):
            raise TypeError(utils.type_error(type(participant), Participant))

        if participant.id in self._participants:
            raise ValueError(utils.duplicate_key_error(self.name, participant.id))

        self._participants._store[participant.id] = participant

    def remove_participant(self, participant: Participant) -> None:
        """Add a resource.

        Args:
            participant: Participant object that is part of the graph.

        Examples:
            >>> g = Graph()
            >>> p1 = Participant("Participant1")
            >>> g.add_participant(p1)
            >>> len(g.participants)
            1
            >>> g.remove_participant(p1)
            >>> len(g.participants)
            0
        """
        if not isinstance(participant, Participant):
            raise TypeError(utils.type_error(type(participant), Participant))

        if participant.id not in self._participants:
            raise ValueError(utils.missing_key_error(self.name, participant.id))

        del self._participants._store[participant.id]

        for node in self._atom_nodes.values():
            if node.participant == participant:
                node._participant = None

    def clean_participants(self) -> None:
        """Remove resources from the graph that are used by no nodes"""

        node_participants = {
            node.participant.id
            for node in self._atom_nodes.values()
            if node.participant
        }

        for participant in set(self._participants):
            if participant not in node_participants:
                del self._participants._store[participant]

    def add_analyst(self, analyst: Analyst) -> None:
        """Add a resource.

        Args:
            analyst: analyst object that is not part of the graph.

        Examples:
            >>> g = Graph()
            >>> p1 = Analyst("Name")
            >>> g.add_analyst(p1)
            >>> len(g.analysts)
            1
        """
        if not isinstance(analyst, Analyst):
            raise TypeError(utils.type_error(type(analyst), Analyst))

        if analyst.id in self._analysts:
            raise ValueError(utils.duplicate_key_error(self.name, analyst.id))

        self._analysts._store[analyst.id] = analyst

    def remove_analyst(self, analyst: Analyst) -> None:
        """Add a resource.

        Args:
            analyst: analyst object that is part of the graph.

        Examples:
            >>> import arguebuf
            >>> g = Graph()
            >>> p1 = Analyst("Name")
            >>> g.add_analyst(p1)
            >>> len(g.analysts)
            1
            >>> g.remove_analyst(p1)
            >>> len(g.analysts)
            0
        """
        if not isinstance(analyst, Analyst):
            raise TypeError(utils.type_error(type(analyst), Analyst))

        if analyst.id not in self._analysts:
            raise ValueError(utils.missing_key_error(self.name, analyst.id))

        del self._analysts._store[analyst.id]

    def strip_scheme_nodes(self) -> None:
        """Remove scheme nodes from graph to connect atom nodes directly

        Can be useful to analyze the structure of atom nodes
        without considering their relation types (i.e., the scheme nodes between them).
        """

        schemes = list(self._scheme_nodes.values())

        for scheme in schemes:
            for incoming, outgoing in itertools.product(
                self.incoming_atom_nodes(scheme), self.outgoing_atom_nodes(scheme)
            ):
                self.add_edge(
                    Edge(
                        incoming,
                        outgoing,
                    )
                )

            self.remove_node(scheme)
